#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Automatically fix lintian issues."""

import logging
import os
import sys
import time
from contextlib import ExitStack
from datetime import datetime
from typing import (
    Iterable,
    List,
    Optional,
    Tuple,
)

import breezy.bzr  # noqa: F401
import breezy.git  # noqa: F401
from breezy.tree import Tree
from breezy.workingtree import WorkingTree
from breezy.workspace import check_clean_tree
from debmutate.reformatting import FormattingUnpreservable

from . import _lintian_brush_rs

__version__ = (0, 152)
version_string = ".".join(map(str, __version__))
SUPPORTED_CERTAINTIES = ["certain", "confident", "likely", "possible", None]
DEFAULT_MINIMUM_CERTAINTY = "certain"
USER_AGENT = "lintian-brush/" + version_string
# Too aggressive?
DEFAULT_URLLIB_TIMEOUT = 3
logger = logging.getLogger(__name__)


DEFAULT_ADDON_FIXERS = _lintian_brush_rs.DEFAULT_ADDON_FIXERS
LintianIssue = _lintian_brush_rs.LintianIssue
FixerResult = _lintian_brush_rs.FixerResult
UnsupportedCertainty = _lintian_brush_rs.UnsupportedCertainty
read_desc_file = _lintian_brush_rs.read_desc_file
only_changes_last_changelog_block = (
    _lintian_brush_rs.only_changes_last_changelog_block)


class NoChanges(Exception):
    """Script didn't make any changes."""

    def __init__(self, fixer, comment=None, overridden_lintian_issues=None):
        super().__init__(fixer, comment)
        self.fixer = fixer
        self.overridden_lintian_issues = overridden_lintian_issues or []


class NotCertainEnough(NoChanges):
    """Script made changes but with too low certainty."""

    def __init__(self, fixer, certainty, minimum_certainty,
                 overridden_lintian_issues=None):
        super().__init__(
            fixer, overridden_lintian_issues=overridden_lintian_issues)
        self.certainty = certainty
        self.minimum_certainty = minimum_certainty


class FixerFailed(Exception):
    """Base class for fixer script failures."""

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.args == other.args


class FixerScriptFailed(FixerFailed):
    """Script failed to run."""

    def __init__(self, path, returncode, errors):
        self.path = path
        self.returncode = returncode
        self.errors = errors

    def __str__(self):
        return "Script %s failed with exit code: %d\n%s\n" % (
            self.path,
            self.returncode,
            self.errors,
        )

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return (
            self.path == other.path
            and self.returncode == other.returncode
            and self.errors == other.errors
        )


class DescriptionMissing(Exception):
    """The fixer script did not provide a description on stdout."""

    def __init__(self, fixer):
        super().__init__(fixer)
        self.fixer = fixer


class NotDebianPackage(Exception):
    """The specified directory does not contain a Debian package."""

    def __init__(self, abspath):
        super().__init__(abspath)


parse_script_fixer_output = _lintian_brush_rs.parse_script_fixer_output
determine_env = _lintian_brush_rs.determine_env
Fixer = _lintian_brush_rs.Fixer
ScriptFixer = _lintian_brush_rs.ScriptFixer
PythonScriptFixer = _lintian_brush_rs.PythonScriptFixer


def open_binary(name):
    return open(data_file_path(name), 'rb')  # noqa: SIM115


def data_file_path(name, check=os.path.exists):
    # There's probably a more Pythonic way of doing this, but
    # I haven't bothered finding out what it is yet..
    path = os.path.abspath(os.path.join(
        os.path.dirname(__file__), "..", name))
    if check(path):
        return path

    import pkg_resources

    path = pkg_resources.resource_filename(
        __name__, f"lintian-brush/{name}")
    if check(path):
        return path

    # Urgh.
    for b in ['/usr/share/lintian-brush',
              '/usr/local/share/lintian-brush',
              os.path.join(sys.prefix, 'share/lintian-brush')]:
        path = os.path.join(b, name)
        if check(path):
            return path
    raise RuntimeError("unable to find data path: %s" % name)


find_fixers_dir = _lintian_brush_rs.find_fixers_dir


def select_fixers(
    fixers: List[Fixer], *, names: Optional[List[str]] = None,
    exclude: Optional[Iterable[str]] = None
) -> List[Fixer]:
    """Select fixers by name, from a list.

    Args:
      fixers: List of Fixer objects
      names: Set of names to select
      exclude: Set of names to exclude
    Raises:
      KeyError: if one of the names did not exist
    """
    select_set = set(names) if names is not None else None
    exclude_set = set(exclude) if exclude is not None else None
    ret = []
    for f in fixers:
        if select_set is not None:
            if f.name not in select_set:
                continue
            select_set.remove(f.name)
        if exclude_set and f.name in exclude_set:
            exclude_set.remove(f.name)
            continue
        ret.append(f)
    if select_set:
        raise KeyError(select_set.pop())
    if exclude_set:
        raise KeyError(exclude_set.pop())
    return ret


def available_lintian_fixers(fixers_dir=None, force_subprocess=False):
    if fixers_dir is None:
        fixers_dir = find_fixers_dir()
    return _lintian_brush_rs.available_lintian_fixers(
        fixers_dir, force_subprocess)


increment_version = _lintian_brush_rs.increment_version


def get_committer(tree: WorkingTree) -> str:
    """Get the committer string for a tree.

    Args:
      tree: A Tree object
    Returns:
      A committer string
    """
    # TODO(jelmer): Perhaps this logic should be in Breezy?
    if hasattr(tree.branch.repository, "_git"):
        cs = tree.branch.repository._git.get_config_stack()
        user = os.environ.get("GIT_COMMITTER_NAME")
        email = os.environ.get("GIT_COMMITTER_EMAIL")
        if user is None:
            try:
                user = cs.get(("user",), "name").decode("utf-8")
            except KeyError:
                user = None
        if email is None:
            try:
                email = cs.get(("user",), "email").decode("utf-8")
            except KeyError:
                email = None
        if user and email:
            return user + " <" + email + ">"
        from breezy.config import GlobalStack

        return GlobalStack().get("email")
    else:
        config = tree.branch.get_config_stack()
        return config.get("email")


certainty_sufficient = _lintian_brush_rs.certainty_sufficient


_changelog_policy_noted = False


def _note_changelog_policy(policy, msg):
    global _changelog_policy_noted
    if not _changelog_policy_noted:
        if policy:
            extra = "Specify --no-update-changelog to override."
        else:
            extra = "Specify --update-changelog to override."
        logging.info("%s %s", msg, extra)
    _changelog_policy_noted = True


class FailedPatchManipulation(Exception):
    def __init__(self, tree, patches_directory, reason):
        super().__init__(
            tree, patches_directory, reason)


def _upstream_changes_to_patch(
    local_tree: WorkingTree,
    basis_tree: Tree,
    dirty_tracker,
    subpath: str,
    patch_name: str,
    patch_description: str,
    timestamp: Optional[datetime] = None,
) -> Tuple[str, List[str]]:
    from .patches import (
        PatchSyntax,
        move_upstream_changes_to_patch,
        read_quilt_patches,
        tree_patches_directory,
    )

    # TODO(jelmer): Apply all patches before generating a diff.

    patches_directory = tree_patches_directory(local_tree, subpath)
    try:
        quilt_patches = list(read_quilt_patches(local_tree, patches_directory))
    except PatchSyntax as e:
        raise FailedPatchManipulation(
            local_tree, patches_directory,
            "Unable to parse some patches: %s" % e) from e
    if len(quilt_patches) > 0:
        raise FailedPatchManipulation(
            local_tree,
            patches_directory,
            "Creating patch on top of existing upstream "
            "patches not supported.",
        )

    logging.debug("Moving upstream changes to patch %s", patch_name)
    try:
        specific_files, patch_name = move_upstream_changes_to_patch(
            local_tree,
            basis_tree,
            subpath,
            patch_name,
            patch_description,
            dirty_tracker,
            timestamp=timestamp,
        )
    except FileExistsError as e:
        raise FailedPatchManipulation(
            local_tree, patches_directory,
            "patch path %s already exists\n" % e.args[0]
        ) from e

    return patch_name, specific_files


run_lintian_fixer = _lintian_brush_rs.run_lintian_fixer


class ManyResult:
    def __init__(self):
        self.success = []
        self.failed_fixers = {}
        self.formatting_unpreservable = {}
        self.overridden_lintian_issues = []
        self.changelog_behaviour = None

    def minimum_success_certainty(self) -> str:
        """Return the minimum certainty of any successfully made change."""
        return min_certainty(
            [
                r.certainty
                for r, unused_summary in self.success
                if r.certainty is not None
            ]
        )


def get_dirty_tracker(
        local_tree: WorkingTree, subpath: str = "",
        use_inotify: Optional[bool] = None):
    """Create a dirty tracker object."""
    if use_inotify is True:
        from breezy.dirty_tracker import DirtyTracker

        return DirtyTracker(local_tree, subpath)
    elif use_inotify is False:
        return None
    else:
        try:
            from breezy.dirty_tracker import DirtyTracker
        except ImportError:
            return None
        else:
            return DirtyTracker(local_tree, subpath)


def determine_update_changelog(local_tree, debian_path):
    from .detect_gbp_dch import (
        ChangelogBehaviour,
        guess_update_changelog,
    )

    changelog_path = os.path.join(debian_path, 'changelog')

    if not local_tree.has_filename(changelog_path):
        # If there's no changelog, then there's nothing to update!
        return False

    behaviour = guess_update_changelog(local_tree, debian_path)
    if behaviour:
        _note_changelog_policy(
            behaviour.update_changelog, behaviour.explanation)
    else:
        # If we can't make an educated guess, assume yes.
        behaviour = ChangelogBehaviour(
            True, "Assuming changelog should be updated")

    return behaviour


def run_lintian_fixers(  # noqa: C901
    local_tree: WorkingTree,
    fixers: List[Fixer],
    update_changelog: bool = True,
    verbose: bool = False,
    committer: Optional[str] = None,
    compat_release: Optional[str] = None,
    minimum_certainty: Optional[str] = None,
    trust_package: bool = False,
    allow_reformatting: bool = False,
    use_inotify: Optional[bool] = None,
    subpath: str = "",
    net_access: bool = True,
    opinionated: Optional[bool] = None,
    diligence: int = 0,
):
    """Run a set of lintian fixers on a tree.

    Args:
      local_tree: WorkingTree object
      fixers: A set of Fixer objects
      update_changelog: Whether to add an entry to the changelog
      verbose: Whether to be verbose
      committer: Optional committer (name and email)
      compat_release: Minimum release that the package should be usable on
        (e.g. 'sid' or 'stretch')
      minimum_certainty: How certain the fixer should be
        about its changes.
      trust_package: Whether to run code from the package if necessary
      allow_reformatting: Whether to allow reformatting of changed files
      use_inotify: Use inotify to watch changes (significantly improves
        performance). Defaults to None (automatic)
      subpath: Subpath in the tree in which the package lives
      net_access: Whether to allow network access
      opinionated: Whether to be opinionated
      diligence: Level of diligence
    Returns:
      Tuple with two lists:
        1. list of tuples with (lintian-tag, certainty, description) of fixers
           that ran
        2. dictionary mapping fixer names for fixers that failed to run to the
           error that occurred
    """
    from tqdm import trange
    basis_tree = local_tree.basis_tree()
    check_clean_tree(local_tree, basis_tree=basis_tree, subpath=subpath)
    fixers = list(fixers)

    # If we don't know whether to update the changelog, then find out *once*
    if update_changelog is None:
        changelog_behaviour = None

        def update_changelog():
            nonlocal update_changelog, changelog_behaviour
            changelog_behaviour = determine_update_changelog(
                local_tree, os.path.join(subpath, "debian"))
            return changelog_behaviour.update_changelog
    else:
        changelog_behaviour = None

    ret = ManyResult()
    with ExitStack() as es:
        t = es.enter_context(
            trange(len(fixers), leave=False, disable=None))  # type: ignore

        dirty_tracker = get_dirty_tracker(
            local_tree, subpath=subpath, use_inotify=use_inotify
        )
        if dirty_tracker:
            from breezy.dirty_tracker import TooManyOpenFiles
            try:
                es.enter_context(dirty_tracker)
            except TooManyOpenFiles:
                logging.warning(
                    'Too many open files for inotify, not using it.')
                dirty_tracker = None

        for fixer in fixers:
            t.set_description("Running fixer %s" % fixer)
            t.update()
            start = time.time()
            if dirty_tracker:
                dirty_tracker.mark_clean()
            try:
                result, summary = run_lintian_fixer(
                    local_tree,
                    fixer,
                    update_changelog=update_changelog,
                    committer=committer,
                    compat_release=compat_release,
                    minimum_certainty=minimum_certainty,
                    trust_package=trust_package,
                    allow_reformatting=allow_reformatting,
                    dirty_tracker=dirty_tracker,
                    subpath=subpath,
                    net_access=net_access,
                    opinionated=opinionated,
                    diligence=diligence,
                    basis_tree=basis_tree,
                )
            except FormattingUnpreservable as e:
                ret.formatting_unpreservable[fixer.name] = e.path
                if verbose:
                    logging.info(
                        "Fixer %r was unable to preserve "
                        "formatting of %s.", fixer.name,
                        e.path)
            except FixerFailed as e:
                ret.failed_fixers[fixer.name] = e
                if verbose:
                    logging.info("Fixer %r failed to run.", fixer.name)
                    sys.stderr.write(str(e))
            except MemoryError as e:
                ret.failed_fixers[fixer.name] = e
                if verbose:
                    logging.info(
                        "Run out of memory while running fixer %r.",
                        fixer.name)
            except NotCertainEnough as e:
                if verbose:
                    logging.info(
                        "Fixer %r made changes but not high enough "
                        "certainty (was %r, needed %r). (took: %.2fs)",
                        fixer.name,
                        e.certainty,
                        e.minimum_certainty,
                        time.time() - start,
                    )
            except FailedPatchManipulation as e:
                if verbose:
                    logging.info(
                        "Unable to manipulate upstream patches: %s",
                        e.args[2])
                ret.failed_fixers[fixer.name] = e
            except NoChanges as e:
                if verbose:
                    logging.info(
                        "Fixer %r made no changes. (took: %.2fs)",
                        fixer.name,
                        time.time() - start,
                    )
                ret.overridden_lintian_issues.extend(
                    e.overridden_lintian_issues)
            else:
                if verbose:
                    logging.info(
                        "Fixer %r made changes. (took %.2fs)",
                        fixer.name,
                        time.time() - start,
                    )
                ret.success.append((result, summary))
                basis_tree = local_tree.basis_tree()
    if changelog_behaviour:
        ret.changelog_behaviour = changelog_behaviour
    return ret


def certainty_to_confidence(certainty: Optional[str]) -> Optional[int]:
    if certainty in ("unknown", None):
        return None
    return SUPPORTED_CERTAINTIES.index(certainty)


min_certainty = _lintian_brush_rs.min_certainty
control_files_in_root = _lintian_brush_rs.control_files_in_root
is_debcargo_package = _lintian_brush_rs.is_debcargo_package
control_file_present = _lintian_brush_rs.control_file_present



#!/usr/bin/python3
# Copyright (C) 2018-2020 Jelmer Vernooij <jelmer@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Functions for working with watch files."""

import json
import logging
import os
import re
import urllib.error
from collections import namedtuple
from contextlib import suppress
from dataclasses import dataclass
from typing import List, Optional, Set
from urllib.parse import urlparse, urlunparse
from urllib.request import Request, urlopen

from debmutate.reformatting import FormattingUnpreservable
from debmutate.versions import get_snapshot_revision, strip_dfsg_suffix
from debmutate.watch import (
    Watch,
    WatchEditor,
    WatchFile,
    apply_url_mangle,
)

from debian.changelog import Changelog

from . import (
    DEFAULT_URLLIB_TIMEOUT,
    USER_AGENT,
    certainty_to_confidence,
    min_certainty,
)
from .gpg import fetch_keys
from .svp import svp_enabled

# TODO(jelmer): Vary this depending on whether a new watch file was added or an
# existing one was fixed?
WATCH_FIX_VALUE = 60


@dataclass
class WatchCandidate:

    watch: Watch
    site: str
    certainty: Optional[str]
    preference: Optional[int]


class KeyRetrievalFailed(Exception):

    def __init__(self, fingerprints):
        self.fingerprints = fingerprints


COMMON_PGPSIGURL_MANGLES = [
    's/$/.%s/' % ext for ext in ['asc', 'pgp', 'gpg', 'sig', 'sign']]


SignatureInfo = namedtuple('SignatureInfo', ['is_valid', 'keys', 'mangle'])


def probe_signature(r, *, pgpsigurlmangle=None, mangles=None,
                    gpg_context=None):
    """Try to find the signature file for a release."""
    import gpg.errors

    if gpg_context is None:
        import gpg
        gpg_context = gpg.Context(armor=True)

    if mangles is None:
        mangles = COMMON_PGPSIGURL_MANGLES

    def sig_valid(sig):
        return sig.status == 0

    if r.pgpsigurl:
        pgpsigurls = [(pgpsigurlmangle, r.pgpsigurl)]
    else:
        pgpsigurls = [
            (mangle, apply_url_mangle(mangle, r.url))
            for mangle in mangles]
    for mangle, pgpsigurl in pgpsigurls:
        # Try and download signatures from some predictable locations.
        try:
            resp = urlopen(pgpsigurl)
        except urllib.error.HTTPError:
            continue
        except TimeoutError:
            logging.warning('Timeout error retrieving %s', pgpsigurl)
            continue
        sig = resp.read()
        actual = urlopen(r.url).read()
        try:
            gr = gpg_context.verify(actual, sig)[1]
        except gpg.errors.GPGMEError as e:
            logging.warning(
                'Error verifying signature %s on %s: %s',
                pgpsigurl, r.url, e)
            continue
        except gpg.errors.BadSignatures as e:
            if str(e).endswith(': No public key'):
                if not fetch_keys(
                        [s.fpr for s in e.result.signatures],
                        home_dir=gpg_context.home_dir):
                    logging.warning(
                        'Unable to retrieve keys: %r',
                        e.result.signatures)
                    raise KeyRetrievalFailed(
                        [s.fpr for s in e.result.signatures]) from e
                gr = gpg_context.verify(actual, sig)[1]
            else:
                raise
        signatures = gr.signatures
        is_valid = True
        needed_keys = set()
        for sig in signatures:
            if not sig_valid(sig):
                logging.warning(
                    'Signature from %s in %s for %s not valid',
                    sig.fpr, pgpsigurl, r.url)
                is_valid = False
            else:
                needed_keys.add(sig.fpr)
        return SignatureInfo(
            is_valid=is_valid, keys=needed_keys, mangle=mangle)
    else:
        return None


def candidates_from_setup_py(
        path, good_upstream_versions: Set[str], net_access=False):
    certainty = "likely"
    # Import setuptools in case it replaces distutils
    with suppress(ImportError):
        import setuptools  # noqa: F401
    from distutils.core import run_setup

    try:
        result = run_setup(os.path.abspath(path), stop_after="config")
    except BaseException:  # noqa: PIE786
        import traceback

        traceback.print_exc()
        return
    project = result.get_name()  # type: ignore
    version = result.get_version()  # type: ignore
    if not project:
        logging.warning('no project name in setup.py')
        return
    current_version_filenames = None
    if net_access:
        # TODO(jelmer): Use ognibuild.upstream.load_pypi
        json_url = "https://pypi.python.org/pypi/%s/json" % project
        logging.info('Getting %s info on pypi (%s)',
                     project, json_url)
        pypi_data = _load_json(json_url)
        if pypi_data is None:
            logging.warning('unable to find project %s on pypi',
                            project)
            return
        if version in pypi_data["releases"]:
            release = pypi_data["releases"][version]
            current_version_filenames = [
                (d["filename"], d["has_sig"])
                for d in release
                if d["packagetype"] == "sdist"
            ]
    filename_regex = (
        fr"{project}-(.+)\.(?:zip|tgz|tbz|txz|(?:tar\.(?:gz|bz2|xz)))")
    opts = []
    # TODO(jelmer): Set uversionmangle?
    # opts.append('uversionmangle=s/(rc|a|b|c)/~$1/')
    if current_version_filenames:
        for (fn, has_sig) in current_version_filenames:
            if re.match(filename_regex, fn):
                certainty = "certain"
                if has_sig:
                    opts.append("pgpsigurlmangle=s/$/.asc/")
    url = fr"https://pypi.debian.net/{project}/{filename_regex}"
    # TODO(jelmer): Add pgpsigurlmangle if has_sig==True
    w = Watch(url, opts=opts)
    yield WatchCandidate(w, "pypi", certainty=certainty, preference=1)


def find_candidates(path, good_upstream_versions, net_access=False):
    candidates = []
    if os.path.exists(os.path.join(path, 'setup.py')):
        candidates.extend(candidates_from_setup_py(
            os.path.join(path, 'setup.py'), good_upstream_versions,
            net_access=net_access))

    if os.path.exists(os.path.join(path, 'debian/upstream/metadata')):
        candidates.extend(candidates_from_upstream_metadata(
            os.path.join(path, 'debian/upstream/metadata'),
            good_upstream_versions, net_access=net_access))

    def candidate_key(candidate):
        return (
            certainty_to_confidence(candidate.certainty),
            candidate.preference)

    candidates.sort(key=candidate_key)

    return candidates


def candidates_from_upstream_metadata(
        path: str, good_upstream_versions: Set[str], net_access: bool = False):
    try:
        with open(path) as f:
            inp = f.read()
    except FileNotFoundError:
        pass
    else:
        import ruamel.yaml

        code = ruamel.yaml.round_trip_load(inp, preserve_quotes=True)

        for field in ["Repository", "X-Download"]:
            try:
                parsed_url = urlparse(code[field].split(' ')[0])
            except KeyError:
                pass
            else:
                if parsed_url.hostname == "github.com":
                    yield from guess_github_watch_entry(
                        parsed_url, good_upstream_versions,
                        net_access=net_access
                    )
                if parsed_url.hostname == "launchpad.net":
                    yield from guess_launchpad_watch_entry(
                        parsed_url, good_upstream_versions,
                        net_access=net_access
                    )

        archive = code.get('Archive')
        if archive == 'CRAN':
            yield from guess_cran_watch_entry(code['Name'])


def guess_cran_watch_entry(name):
    w = Watch(r'https://cran.r-project.org/src/contrib/%s_([-\d.]*)\.tar\.gz'
              % name)
    yield WatchCandidate(w, "cran", certainty="likely", preference=0)


def guess_launchpad_watch_entry(
        parsed_url, good_upstream_versions, net_access=False):
    if not net_access:
        return
    project = parsed_url.path.strip("/").split("/")[0]
    url = f"https://api.launchpad.net/devel/{project}/releases"
    entries = []
    while url:
        response = _load_json(url)
        entries.extend(response['entries'])
        url = response.get('next_collection_link')
    files = _load_json(entries[-1]['files_collection_link'])
    assert len(files['entries']) > 0
    # TODO(jelmer): add
    filepattern = files['entries'][0]['file_link'].split('/')[-2].replace(
        entries[-1]['version'], '(.*)')
    w = Watch(
        f'https://launchpad.net/{project}/+download',
        f'https://launchpad.net/{project}/.*/{filepattern}')
    yield WatchCandidate(w, "launchpad", certainty="certain", preference=0)


def guess_github_watch_entry(
        parsed_url, good_upstream_versions, net_access=False):
    import re

    from breezy.branch import Branch

    if not net_access:
        return
    branch = Branch.open(urlunparse(parsed_url))
    tags = branch.tags.get_tag_dict()
    POSSIBLE_PATTERNS = [r"v(\d\S+)", r"(\d\S+)",
                         r".*/[vV]?(\d[^\s+]+)\.tar\.gz"]
    version_pattern = None
    # TODO(jelmer): Maybe use releases API instead?
    # TODO(jelmer): Automatically added mangling for
    # e.g. rc and beta
    uversionmangle: List[str] = []
    for name in sorted(tags, reverse=True):
        for pattern in POSSIBLE_PATTERNS:
            m = re.match(pattern, name)
            if not m:
                continue
            if m.group(1) in good_upstream_versions:
                version_pattern = pattern
                break
        if version_pattern:
            break
    else:
        return
    (username, project) = parsed_url.path.strip("/").split("/")
    if project.endswith(".git"):
        project = project[:-4]
    download_url = f"https://github.com/{username}/{project}/tags"
    matching_pattern = r".*\/%s\.tar\.gz" % version_pattern
    opts = [
        fr"filenamemangle=s/{matching_pattern}/{project}-$1\.tar\.gz/"
    ]
    if uversionmangle:
        opts.append(r"uversionmangle=" + ";".join(uversionmangle))
    # TODO(jelmer): Check for GPG
    # opts.append(
    #    r'pgpsigurlmangle='
    #    r's/archive\/%s\.tar\.gz/releases\/download\/$1\/$1\.tar\.gz\.asc/' %
    #    version_pattern)
    w = Watch(download_url, matching_pattern, opts=opts)
    yield WatchCandidate(w, "github", certainty="certain", preference=0)


def _load_json(url):
    headers = {"User-Agent": USER_AGENT, "Accept": "application/json"}
    try:
        response = urlopen(
            Request(url, headers=headers), timeout=DEFAULT_URLLIB_TIMEOUT
        )
    except urllib.error.HTTPError as e:
        if e.status == 404:
            return None
        raise
    return json.load(response)


def candidates_from_hackage(package, good_upstream_versions, net_access=False):
    if not net_access:
        return
    url = "https://hackage.haskell.org/package/%s/preferred" % package
    versions = _load_json(url)
    if versions is None:
        return
    for version in versions["normal-version"]:
        if version in good_upstream_versions:
            break
    else:
        return
    download_url = "https://hackage.haskell.org/package/" + package
    matching_pattern = r".*/%s-(.*).tar.gz" % package
    w = Watch(download_url, matching_pattern)
    yield WatchCandidate(w, "hackage", certainty="certain", preference=1)


def fix_old_github_patterns(updater):
    ret = []
    for w in getattr(updater.watch_file, "entries", []):
        parsed_url = urlparse(w.url)

        # only applies to github.com
        if parsed_url.netloc != "github.com":
            continue

        parts = parsed_url.path.strip('/').split('/')
        if len(parts) >= 3 and parts[2] in ('tags', 'releases'):  # noqa:SIM114
            pass
        elif len(parts) >= 2 and parts[0] == '.*' and parts[1] == 'archive':
            pass
        else:
            continue

        parts = w.matching_pattern.split('/')
        if len(parts) > 2 and parts[-2] == 'archive':
            parts.insert(-1, 'refs/tags')
        w.matching_pattern = '/'.join(parts)
        ret.append(w)
    return ret


def fix_github_releases(updater):
    ret = []
    for w in getattr(updater.watch_file, "entries", []):
        parsed_url = urlparse(w.url)

        # only applies to github.com
        if parsed_url.netloc != "github.com":
            continue

        parts = parsed_url.path.strip('/').split('/')
        if len(parts) >= 3 and parts[2] == 'releases':
            parts[2] = 'tags'
            parsed_url = parsed_url._replace(path='/'.join(parts))
            w.url = parsed_url.geturl()
            ret.append(w)
    return ret


def fix_watch_issues(updater):
    ret = []
    ret.extend(fix_old_github_patterns(updater))
    ret.extend(fix_github_releases(updater))
    return ret


def watch_entries_certainty(entries, source_package,
                            expected_versions=None,
                            default_certainty="likely"):
    certainty = "certain"
    for entry in entries:
        try:
            ret = verify_watch_entry(
                entry, source_package,
                expected_versions=expected_versions)
        except WatchEntryVerificationFailure:
            certainty = min_certainty(["possible", certainty])
        except TemporaryWatchEntryVerficationError:
            certainty = min_certainty([default_certainty, certainty])
        else:
            if not ret:
                certainty = min_certainty(["possible", certainty])
    return certainty


class WatchEntryVerificationFailure(Exception):
    """Failure verifying watch entry."""


class TemporaryWatchEntryVerficationError(Exception):
    """Temporary error verifying watch entry."""


class WatchEntryVerificationStatus:

    def __init__(self, entry, releases, missing_versions=None):
        self.entry = entry
        self.releases = {r.version: r for r in releases}
        self.missing_versions = missing_versions

    def __bool__(self):
        return not self.missing_versions and bool(self.releases)


def verify_watch_entry(
        entry: Watch, source_package: str,
        expected_versions: Optional[
            List[str]] = None) -> WatchEntryVerificationStatus:
    try:
        releases = list(sorted(
                entry.discover(source_package), reverse=True))
    except urllib.error.HTTPError as e:
        logging.warning(
            'HTTP error accessing discovery URL %s: %s.',
            e.geturl(), e)
        if (e.status or 0) // 100 == 5:
            # If the server is unhappy, then the entry could still be valid.
            raise TemporaryWatchEntryVerficationError(str(e)) from e

        raise WatchEntryVerificationFailure(str(e)) from e

    if expected_versions is None:
        return WatchEntryVerificationStatus(entry, releases=releases)

    found_versions = {str(r.version) for r in releases}
    missing_versions = set(expected_versions) - found_versions
    return WatchEntryVerificationStatus(
        entry, releases=releases, missing_versions=missing_versions)


def report_fatal(code: str, description: str, context=None, hint=None) -> None:
    if os.environ.get('SVP_API') == '1':
        with open(os.environ['SVP_RESULT'], 'w') as f:
            json.dump({
                'result_code': code,
                'description': description,
                'context': context}, f)
    logging.fatal('%s', description)
    if hint:
        logging.info('%s', hint)


def verify_watch_file(watch_file, source_package, expected_versions):
    ret = []
    for entry in getattr(watch_file, 'entries', []):
        ret.append(verify_watch_entry(
            entry, source_package, expected_versions))
    return ret


def svp_context(status, site):
    return [{
        'entries': {
            'text': str(entry_status.entry),
            'site': entry_site,
            'releases': {
                str(r.version): {
                    'url': r.url,
                } for r in entry_status.releases.values()
            },
            'missing_versions':
                [str(x) for x in entry_status.missing_versions]
        }
    } for entry_status, entry_site in zip(
        status or [], site or [None] * len(status or []))]


_changelog_policy_noted = False


def _note_changelog_policy(policy, msg):
    global _changelog_policy_noted
    if not _changelog_policy_noted:
        if policy:
            extra = "Specify --no-update-changelog to override."
        else:
            extra = "Specify --update-changelog to override."
        logging.info("%s %s", msg, extra)
    _changelog_policy_noted = True


def main():  # noqa: C901
    import argparse
    import logging

    import breezy  # noqa: E402

    breezy.initialize()  # type: ignore
    import breezy.bzr  # noqa: E402
    import breezy.git  # noqa: E402
    from breezy.commit import NullCommitReporter, PointlessCommit
    from breezy.workingtree import WorkingTree
    from breezy.workspace import (
        WorkspaceDirty,
        check_clean_tree,
    )

    from . import (
        get_committer,
        version_string,
    )
    from .config import Config

    parser = argparse.ArgumentParser(prog="deb-update-watch")
    parser.add_argument(
        "--directory",
        metavar="DIRECTORY",
        help="directory to run in",
        type=str,
        default=".",
    )
    parser.add_argument(
        "--no-update-changelog",
        action="store_false",
        default=None,
        dest="update_changelog",
        help="do not update the changelog",
    )
    parser.add_argument(
        "--update-changelog",
        action="store_true",
        dest="update_changelog",
        help="force updating of the changelog",
        default=None,
    )
    parser.add_argument(
        "--allow-reformatting",
        default=None,
        action="store_true",
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--version", action="version", version="%(prog)s " + version_string
    )
    parser.add_argument(
        "--identity",
        help="Print user identity that would be used when committing",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--debug", help="Describe all considered changes.", action="store_true"
    )
    parser.add_argument(
        "--force",
        help="Force regenerating watch file, even if it the current one works",
        action="store_true")
    parser.add_argument(
        "--no-verify", action="store_true",
        help="Do not verify that the new watch file works")
    parser.add_argument(
        "--disable-net-access",
        help="Do not probe external services.",
        action="store_true",
        default=False,
    )

    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO, format='%(message)s')

    wt, subpath = WorkingTree.open_containing(args.directory)
    if args.identity:
        logging.info('%s', get_committer(wt))
        return 0

    with wt.lock_write():
        try:
            check_clean_tree(wt, wt.basis_tree(), subpath)
        except WorkspaceDirty:
            logging.info("%s: Please commit pending changes first.",
                         wt.basedir)
            return 1

        update_changelog = args.update_changelog
        allow_reformatting = args.allow_reformatting
        try:
            cfg = Config.from_workingtree(wt, subpath)
        except FileNotFoundError:
            pass
        else:
            if update_changelog is None:
                update_changelog = cfg.update_changelog()
            if allow_reformatting is None:
                allow_reformatting = cfg.allow_reformatting()

        if allow_reformatting is None:
            allow_reformatting = False

        good_upstream_versions = set()

        changelog_path = os.path.join(subpath, "debian/changelog")

        with wt.get_file(changelog_path) as f:
            cl = Changelog(f, max_blocks=1)
            for entry in cl:
                uv = strip_dfsg_suffix(entry.version.upstream_version)
                if get_snapshot_revision(uv) is not None:
                    continue
                good_upstream_versions.add(uv)
            package = cl.package

        expected_versions = list(sorted(good_upstream_versions))[-5:]

        status = None
        try:
            with WatchEditor(allow_reformatting=allow_reformatting,
                             allow_missing=False,
                             ) as updater:
                if not args.force:
                    try:
                        status = verify_watch_file(
                            updater.watch_file, package, expected_versions)
                    except (TemporaryWatchEntryVerficationError,
                            WatchEntryVerificationFailure):
                        status = [None]
                    if status and all(status):
                        hint = (
                            'Releases {} can be found with watch entry {}'
                            .format(
                                ', '.join(sorted(list(
                                    status[0].releases.keys()), reverse=True)),
                                status[0].entry))
                        report_fatal(
                            'nothing-to-do',
                            'Existing watch file has valid entries',
                            context=svp_context(status, site=None),
                            hint=hint)
                        return 0
                    fix_watch_issues(updater)
                    try:
                        status = verify_watch_file(
                            updater.watch_file, package, expected_versions)
                    except WatchEntryVerificationFailure:
                        status = [None]
                    except TemporaryWatchEntryVerficationError as e:
                        report_fatal('temporary-verification-error', str(e))
                        return 1
                if status is None or not all(status):
                    candidates = find_candidates(
                        '.', good_upstream_versions,
                        net_access=not args.disable_net_access)
                    if len(candidates) == 0:
                        report_fatal(
                            'no-download-urls-found',
                            'No candidates for the watch file were found')
                        return 1
                    updater.allow_reformatting = True
                    updater.watch_file.entries = [candidates[0].watch]
                    site = [candidates[0].site]
                    status = None
                    summary = f'Update watch file from {site[0]}'
                else:
                    site = None
                    summary = 'Fixed watch file'
            specific_files = updater.changed_files
        except FileNotFoundError:
            candidates = find_candidates(
                '.', good_upstream_versions,
                net_access=not args.disable_net_access)
            wf = WatchFile()
            if len(candidates) == 0:
                report_fatal(
                    'no-download-urls-found',
                    'No candidates for the watch file were found')
                return 1
            wf.entries.append(candidates[0].watch)
            site = [candidates[0].site]

            with open('debian/watch', 'w') as f:
                wf.dump(f)
            status = None
            summary = f'Added watch file from {site}'
            specific_files = ['debian/watch']
        except FormattingUnpreservable as e:
            report_fatal('formatting-unpreservable',
                         "Unable to preserve formatting of %s" % e.path)
            if hasattr(e, 'diff'):  # debmutate >= 0.64
                sys.stderr.writelines(e.diff())
            return 1

        if not args.no_verify and status is None:
            with WatchEditor() as updater:
                try:
                    status = verify_watch_file(
                        updater.watch_file, package, expected_versions)
                except WatchEntryVerificationFailure as e:
                    report_fatal(
                        'verification-failed',
                        'Unable to verify watch entry: %s'
                        % e)
                    return 1
                except TemporaryWatchEntryVerficationError as e:
                    report_fatal(
                        'temporary-verification-error',
                        'Unable to verify watch entry: %s'
                        % e)
                    return 1
                if not status or not all(status):
                    report_fatal(
                        'verification-failed',
                        'Unable to watch entries; missing versions: %r'
                        % status[0].missing_versions,
                        context=svp_context(status, site))
                    return 1

        if update_changelog is None:
            from .detect_gbp_dch import guess_update_changelog

            dch_guess = guess_update_changelog(
                wt, os.path.join(subpath, 'debian'))
            if dch_guess:
                update_changelog = dch_guess.update_changelog
                _note_changelog_policy(update_changelog, dch_guess.explanation)
            else:
                # Assume we should update changelog
                update_changelog = True

        if update_changelog:
            from .changelog import add_changelog_entry

            # TODO(jelmer): Add note here about having verified watch file?
            add_changelog_entry(wt, changelog_path, [summary])
            specific_files.append(changelog_path)

        committer = get_committer(wt)

        with suppress(PointlessCommit):
            wt.commit(
                specific_files=specific_files,
                message=f"{summary}\n\nChanges-By: fix-watch-file",
                allow_pointless=False,
                reporter=NullCommitReporter(),
                committer=committer,
            )

    if svp_enabled():
        with open(os.environ["SVP_RESULT"], "w") as f:
            json.dump({
                "description": summary,
                "context": svp_context(status, site),
                "value": WATCH_FIX_VALUE,
            }, f)

    logging.info('%s', summary)

    if status is not None:
        logging.info('Releases {} can be found with new watch entry {}'.format(
            ', '.join(sorted(list(
                status[0].releases.keys()), reverse=True)),
            status[0].entry))

    return 0


if __name__ == '__main__':
    import sys

    sys.exit(main())



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for lintian_brush."""

import os
import re
import sys
from datetime import datetime
from typing import Any, Callable, Type

from breezy.config import GlobalStack
from breezy.tests import (
    TestCase,
    TestCaseWithTransport,
)

from debian.changelog import (
    Changelog,
)
from debian.debian_support import Version
from lintian_brush import (
    FailedPatchManipulation,
    FixerFailed,
    FixerResult,
    FixerScriptFailed,
    NoChanges,
    NotDebianPackage,
    PythonScriptFixer,
    ScriptFixer,
    UnsupportedCertainty,
    available_lintian_fixers,
    certainty_sufficient,
    certainty_to_confidence,
    get_committer,
    increment_version,
    min_certainty,
    only_changes_last_changelog_block,
    parse_script_fixer_output,
    run_lintian_fixer,
    run_lintian_fixers,
    select_fixers,
    version_string,
)

CHANGELOG_FILE = (
    "debian/changelog",
    """\
blah (%(version)s) UNRELEASED; urgency=medium

  * Initial release. (Closes: #911016)

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100
""",
)


class AvailableLintianFixersTest(TestCaseWithTransport):
    def test_find_shell_scripts(self):
        self.build_tree(
            [
                "fixers/",
                "fixers/anotherdir/",
                "fixers/i-fix-a-tag.sh",
                "fixers/i-fix-another-tag.py",
                "fixers/.hidden",
                "fixers/backup-file.sh~",
                "fixers/no-extension",
            ]
        )
        self.build_tree_contents(
            [
                (
                    "fixers/index.desc",
                    """\
- script: foo.sh
  lintian-tags:
   - i-fix-a-tag

- script: bar.sh
  lintian-tags:
   - i-fix-another-tag
   - no-extension
""",
                )
            ]
        )
        self.assertEqual(
            [["i-fix-a-tag"], ["i-fix-another-tag", "no-extension"]],
            [fixer.lintian_tags
             for fixer in available_lintian_fixers("fixers")],
        )


class DummyFixer:
    def __init__(self, name, tags):
        self.name = name
        self.lintian_tags = tags

    def run(self, basedir, package, *args, **kwargs):
        with open(os.path.join(basedir, "debian/control"), "a") as f:
            f.write("a new line\n")
        return FixerResult(
            "Fixed some tag.\nExtended description.", ["some-tag"], "certain"
        )


class FailingFixer:

    def __init__(self, name, tags):
        self.name = name
        self.lintian_tags = tags

    def run(self, basedir, package, *args, **kwargs):
        with open(os.path.join(basedir, "debian/foo"), "w") as f:
            f.write("blah")
        with open(os.path.join(basedir, "debian/control"), "a") as f:
            f.write("foo\n")
        raise FixerFailed("Not successful")


class RunLintianFixerTests(TestCaseWithTransport):
    def make_test_tree(self, format=None, version="0.1"):
        tree = self.make_branch_and_tree(".", format=format)
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/control",
                    """\
Source: blah
Vcs-Git: https://example.com/blah
Testsuite: autopkgtest

Binary: blah
Arch: all

""",
                ),
                (CHANGELOG_FILE[0], CHANGELOG_FILE[1] % {"version": version}),
            ]
        )
        tree.add(["debian", "debian/changelog", "debian/control"])
        tree.commit("Initial thingy.")
        return tree

    def test_not_debian_tree(self):
        tree = self.make_test_tree()
        tree.remove("debian/changelog")
        os.remove("debian/changelog")
        tree.commit("not a debian dir")
        with tree.lock_write():
            self.assertRaises(
                NotDebianPackage,
                run_lintian_fixer,
                tree,
                DummyFixer("dummy", ["some-tag"]),
                update_changelog=False,
            )

    def test_simple_modify(self):
        tree = self.make_test_tree()
        with tree.lock_write():
            result, summary = run_lintian_fixer(
                tree, DummyFixer("dummy", ["some-tag"]), update_changelog=False
            )
        self.assertEqual(summary, "Fixed some tag.")
        self.assertEqual(["some-tag"], result.fixed_lintian_tags)
        self.assertEqual("certain", result.certainty)
        self.assertEqual(2, tree.branch.revno())
        self.assertEqual(
            tree.get_file_lines("debian/control")[-1],
            b"a new line\n")

    def test_simple_modify_too_uncertain(self):
        tree = self.make_test_tree()

        class UncertainFixer:
            def __init__(self, name, tags):
                self.name = name
                self.lintian_tags = tags

            def run(self, basedir, package, *args, **kwargs):
                with open(os.path.join(basedir, "debian/somefile"), "w") as f:
                    f.write("test")
                return FixerResult("Renamed a file.", certainty="possible")

        with tree.lock_write():
            self.assertRaises(
                NoChanges,
                run_lintian_fixer,
                tree,
                UncertainFixer("dummy", ["some-tag"]),
                update_changelog=False,
                minimum_certainty="certain",
            )
        self.assertEqual(1, tree.branch.revno())

    def test_simple_modify_acceptably_uncertain(self):
        tree = self.make_test_tree()

        class UncertainFixer:
            def __init__(self, name, tags):
                self.name = name
                self.lintian_tags = tags

            def run(self, basedir, package, *args, **kwargs):
                with open(os.path.join(basedir, "debian/somefile"), "w") as f:
                    f.write("test")
                return FixerResult("Renamed a file.", certainty="possible")

        with tree.lock_write():
            result, summary = run_lintian_fixer(
                tree,
                UncertainFixer("dummy", ["some-tag"]),
                update_changelog=False,
                minimum_certainty="possible",
            )
        self.assertEqual(2, tree.branch.revno())

    def test_new_file(self):
        tree = self.make_test_tree()

        class NewFileFixer:
            def __init__(self, name, tags):
                self.name = name
                self.lintian_tags = tags

            def run(self, basedir, package, *args, **kwargs):
                with open(os.path.join(basedir, "debian/somefile"), "w") as f:
                    f.write("test")
                return FixerResult("Created new file.", ["some-tag"])

        with tree.lock_write():
            result, summary = run_lintian_fixer(
                tree,
                NewFileFixer("new-file", ["some-tag"]), update_changelog=False
            )
        self.assertEqual(summary, "Created new file.")
        self.assertIs(None, result.certainty)
        self.assertEqual(["some-tag"], result.fixed_lintian_tags)
        rev = tree.branch.repository.get_revision(tree.last_revision())
        self.assertEqual(
            rev.message,
            (
                "Created new file.\n"
                "\n"
                "Changes-By: lintian-brush\n"
                "Fixes: lintian: some-tag\n"
                "See-also: https://lintian.debian.org/tags/some-tag.html\n"
            ),
        )
        self.assertEqual(2, tree.branch.revno())
        basis_tree = tree.branch.basis_tree()
        with basis_tree.lock_read():
            self.assertEqual(
                basis_tree.get_file_text("debian/somefile"), b"test")

    def test_rename_file(self):
        tree = self.make_test_tree()

        class RenameFileFixer:
            def __init__(self, name, tags):
                self.name = name
                self.lintian_tags = tags

            def run(self, basedir, package, *args, **kwargs):
                os.rename(
                    os.path.join(basedir, "debian/control"),
                    os.path.join(basedir, "debian/control.blah"),
                )
                return FixerResult("Renamed a file.")

        orig_basis_tree = tree.branch.basis_tree()
        with tree.lock_write():
            result, summary = run_lintian_fixer(
                tree, RenameFileFixer("rename", ["some-tag"]),
                update_changelog=False
            )
        self.assertEqual(summary, "Renamed a file.")
        self.assertIs(None, result.certainty)
        self.assertEqual([], result.fixed_lintian_tags)
        self.assertEqual(2, tree.branch.revno())
        basis_tree = tree.branch.basis_tree()
        with basis_tree.lock_read(), orig_basis_tree.lock_read():
            self.assertFalse(basis_tree.has_filename("debian/control"))
            self.assertTrue(basis_tree.has_filename("debian/control.blah"))
            self.assertNotEqual(
                orig_basis_tree.get_revision_id(), basis_tree.get_revision_id()
            )
            self.expectFailure(
                "mv --auto appears to be broken",
                self.assertEqual,
                basis_tree.path2id("debian/control.blah"),
                orig_basis_tree.path2id("debian/control"),
            )

    def test_empty_change(self):
        tree = self.make_test_tree()

        class EmptyFixer:
            def __init__(self, name, tags):
                self.name = name
                self.lintian_tags = tags

            def run(self, basedir, package, *args, **kwargs):
                return FixerResult("I didn't actually change anything.")

        with tree.lock_write():
            self.assertRaises(
                NoChanges,
                run_lintian_fixer,
                tree,
                EmptyFixer("empty", ["some-tag"]),
                update_changelog=False,
            )
        self.assertEqual(1, tree.branch.revno())
        with tree.lock_read():
            self.assertEqual([], list(tree.iter_changes(tree.basis_tree())))

    def test_fails(self):
        tree = self.make_test_tree()
        with tree.lock_write():
            self.assertRaises(
                Exception,
                run_lintian_fixer,
                tree,
                FailingFixer("fail", ["some-tag"]),
                update_changelog=False,
            )
        self.assertEqual(1, tree.branch.revno())
        with tree.lock_read():
            self.assertEqual([], list(tree.iter_changes(tree.basis_tree())))

    def test_upstream_change(self):
        tree = self.make_test_tree(version="0.1-1")

        class NewFileFixer:
            def __init__(self, name, tags):
                self.name = name
                self.lintian_tags = tags

            def run(self, basedir, package, *args, **kwargs):
                with open(os.path.join(basedir, "configure.ac"), "w") as f:
                    f.write("AC_INIT(foo, bar)\n")
                return FixerResult(
                    "Created new configure.ac.", [], patch_name="add-config"
                )

        with tree.lock_write():
            result, summary = run_lintian_fixer(
                tree,
                NewFileFixer("add-config", ["add-config"]),
                update_changelog=False,
                timestamp=datetime(2020, 9, 8, 0, 36, 35, 857836),
            )
        self.assertEqual(
            summary, "Add patch add-config.patch: Created new configure.ac."
        )
        self.assertIs(None, result.certainty)
        self.assertEqual([], result.fixed_lintian_tags)
        rev = tree.branch.repository.get_revision(tree.last_revision())
        self.assertEqual(
            rev.message,
            ("Created new configure.ac.\n" "\n" "Changes-By: lintian-brush\n"),
        )
        self.assertEqual(2, tree.branch.revno())
        basis_tree = tree.branch.basis_tree()
        with basis_tree.lock_read():
            self.assertEqual(
                basis_tree.get_file_text("debian/patches/series"),
                b"add-config.patch\n"
            )
            lines = basis_tree.get_file_lines(
                "debian/patches/add-config.patch")
        self.assertEqual(lines[0], b"Description: Created new configure.ac.\n")
        self.assertEqual(lines[1], b"Origin: other\n")
        self.assertEqual(lines[2], b"Last-Update: 2020-09-08\n")
        self.assertEqual(lines[3], b"\n")
        self.assertEqual(lines[4], b"=== added file 'configure.ac'\n")
        self.assertEqual(lines[7], b"@@ -0,0 +1,1 @@\n")
        self.assertEqual(lines[8], b"+AC_INIT(foo, bar)\n")

    def test_upstream_change_stacked(self):
        tree = self.make_test_tree(version="0.1-1")

        self.build_tree_contents(
            [
                ("debian/patches/",),
                ("debian/patches/series", "foo\n"),
                (
                    "debian/patches/foo",
                    """\
--- /dev/null	2020-09-07 13:26:27.546468905 +0000
+++ a	2020-09-08 01:26:25.811742671 +0000
@@ -0,0 +1 @@
+foo
""",
                ),
            ]
        )
        tree.add([
            "debian/patches",
            "debian/patches/series",
            "debian/patches/foo"
        ])
        tree.commit("Add patches")

        class NewFileFixer:
            def __init__(self, name, tags):
                self.name = name
                self.lintian_tags = tags

            def run(self, basedir, package, *args, **kwargs):
                with open(os.path.join(basedir, "configure.ac"), "w") as f:
                    f.write("AC_INIT(foo, bar)\n")
                return FixerResult(
                    "Created new configure.ac.", [], patch_name="add-config"
                )

        with tree.lock_write():
            self.assertRaises(
                FailedPatchManipulation,
                run_lintian_fixer,
                tree,
                NewFileFixer("add-config", ["add-config"]),
                update_changelog=False,
                timestamp=datetime(2020, 9, 8, 0, 36, 35, 857836),
            )


class RunLintianFixersTests(TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        self.tree = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/control",
                    """\
Source: blah
Vcs-Git: https://example.com/blah
Testsuite: autopkgtest

Binary: blah
Arch: all

""",
                ),
                (CHANGELOG_FILE[0], CHANGELOG_FILE[1] % {"version": "0.1"}),
            ]
        )
        self.tree.add(["debian", "debian/changelog", "debian/control"])
        self.tree.commit("Initial thingy.")

    def test_fails(self):
        with self.tree.lock_write():
            result = run_lintian_fixers(
                self.tree, [FailingFixer("fail", ["some-tag"])],
                update_changelog=False
            )
        self.assertEqual([], result.success)
        self.assertEqual(
            {"fail": FixerFailed("Not successful")},
            result.failed_fixers)
        with self.tree.lock_read():
            self.assertEqual(
                [],
                list(self.tree.iter_changes(self.tree.basis_tree())))

    def test_not_debian_tree(self):
        self.tree.remove("debian/changelog")
        os.remove("debian/changelog")
        self.tree.commit("not a debian dir")
        with self.tree.lock_write():
            self.assertRaises(
                NotDebianPackage,
                run_lintian_fixers,
                self.tree,
                [DummyFixer("dummy", ["some-tag"])],
                update_changelog=False,
            )

    def test_simple_modify(self):
        with self.tree.lock_write():
            result = run_lintian_fixers(
                self.tree, [DummyFixer("dummy", ["some-tag"])],
                update_changelog=False
            )
            revid = self.tree.last_revision()
        self.assertEqual(
            [
                (
                    FixerResult(
                        "Fixed some tag.\nExtended description.",
                        ["some-tag"],
                        "certain",
                        revision_id=revid,
                    ),
                    "Fixed some tag.",
                )
            ],
            result.success,
        )
        self.assertEqual({}, result.failed_fixers)
        self.assertEqual(2, self.tree.branch.revno())
        self.assertEqual(
            self.tree.get_file_lines("debian/control")[-1], b"a new line\n"
        )


class HonorsVcsCommitter(TestCaseWithTransport):
    def make_package_tree(self, format, version="0.1"):
        tree = self.make_branch_and_tree(".", format=format)
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/control",
                    """\
Source: blah
Vcs-Git: https://example.com/blah
Testsuite: autopkgtest

Binary: blah
Arch: all

""",
                ),
                (CHANGELOG_FILE[0], CHANGELOG_FILE[1] % {"version": version}),
            ]
        )
        tree.add(["debian", "debian/changelog", "debian/control"])
        tree.commit("Initial thingy.")
        return tree

    def make_change(self, tree, committer=None):
        with tree.lock_write():
            result, summary = run_lintian_fixer(
                tree,
                DummyFixer("dummy", ["some-tag"]),
                update_changelog=False,
                committer=committer,
            )
        self.assertEqual(summary, "Fixed some tag.")
        self.assertEqual(["some-tag"], result.fixed_lintian_tags)
        self.assertEqual("certain", result.certainty)
        self.assertEqual(2, tree.branch.revno())
        self.assertEqual(
            tree.get_file_lines("debian/control")[-1],
            b"a new line\n")

    def test_honors_tree_committer_specified(self):
        tree = self.make_package_tree("git")
        self.make_change(tree, committer="Jane Example <jane@example.com>")
        rev = tree.branch.repository.get_revision(tree.branch.last_revision())
        self.assertEqual(rev.committer, "Jane Example <jane@example.com>")

    def test_honors_tree_committer_config(self):
        tree = self.make_package_tree("git")
        with open(os.path.join(tree.basedir, ".git/config"), "w") as f:
            f.write(
                """\
[user]
  email = jane@example.com
  name = Jane Example

"""
            )

        self.make_change(tree)
        rev = tree.branch.repository.get_revision(tree.branch.last_revision())
        self.assertEqual(rev.committer, "Jane Example <jane@example.com>")


class IncrementVersionTests(TestCase):
    def assertVersion(self, expected, start):
        v = Version(start)
        v = increment_version(v)
        self.assertEqual(Version(expected), v)

    def test_full(self):
        self.assertVersion("1.0-2", "1.0-1")

    def test_native(self):
        self.assertVersion("1.1", "1.0")


class OnlyChangesLastChangelogBlockTests(TestCaseWithTransport):
    def make_package_tree(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/control",
                    """\
Source: blah
Vcs-Git: https://example.com/blah
Testsuite: autopkgtest

Binary: blah
Arch: all

""",
                ),
                (
                    "debian/changelog",
                    """\
blah (0.2) UNRELEASED; urgency=medium

  * And a change.

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100

blah (0.1) unstable; urgency=medium

  * Initial release. (Closes: #911016)

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100
""",
                ),
            ]
        )
        tree.add(["debian", "debian/changelog", "debian/control"])
        tree.commit("Initial thingy.")
        return tree

    def test_no_changes(self):
        tree = self.make_package_tree()
        basis_tree = tree.basis_tree()
        with tree.lock_read(), basis_tree.lock_read():
            changes = tree.iter_changes(basis_tree)
        self.assertFalse(
            only_changes_last_changelog_block(
                tree, tree.basis_tree(), "debian/changelog", list(changes)
            )
        )

    def test_other_change(self):
        tree = self.make_package_tree()
        self.build_tree_contents(
            [
                (
                    "debian/control",
                    """\
Source: blah
Vcs-Bzr: https://example.com/blah
Testsuite: autopkgtest

Binary: blah
Arch: all

""",
                )
            ]
        )
        basis_tree = tree.basis_tree()
        with tree.lock_read(), basis_tree.lock_read():
            changes = tree.iter_changes(basis_tree)
        self.assertFalse(
            only_changes_last_changelog_block(
                tree, tree.basis_tree(), "debian/changelog", list(changes)
            )
        )

    def test_other_changes(self):
        tree = self.make_package_tree()
        self.build_tree_contents(
            [
                (
                    "debian/control",
                    """\
Source: blah
Vcs-Bzr: https://example.com/blah
Testsuite: autopkgtest

Binary: blah
Arch: all

""",
                ),
                (
                    "debian/changelog",
                    """\
blah (0.1) UNRELEASED; urgency=medium

  * Initial release. (Closes: #911016)
  * Some other change.

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100
""",
                ),
            ]
        )
        basis_tree = tree.basis_tree()
        with tree.lock_read(), basis_tree.lock_read():
            changes = tree.iter_changes(basis_tree)
        self.assertFalse(
            only_changes_last_changelog_block(
                tree, tree.basis_tree(), "debian/changelog", list(changes)
            )
        )

    def test_changes_to_other_changelog_entries(self):
        tree = self.make_package_tree()
        self.build_tree_contents(
            [
                (
                    "debian/changelog",
                    """\
blah (0.2) UNRELEASED; urgency=medium

  * debian/changelog: And a change.

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100

blah (0.1) unstable; urgency=medium

  * debian/changelog: Initial release. (Closes: #911016)

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100
""",
                )
            ]
        )
        basis_tree = tree.basis_tree()
        with tree.lock_read(), basis_tree.lock_read():
            changes = tree.iter_changes(basis_tree)
        self.assertFalse(
            only_changes_last_changelog_block(
                tree, tree.basis_tree(), "debian/changelog", list(changes)
            )
        )

    def test_changes_to_last_only(self):
        tree = self.make_package_tree()
        self.build_tree_contents(
            [
                (
                    "debian/changelog",
                    """\
blah (0.2) UNRELEASED; urgency=medium

  * And a change.
  * Not a team upload.

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100

blah (0.1) unstable; urgency=medium

  * Initial release. (Closes: #911016)

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100
""",
                )
            ]
        )
        basis_tree = tree.basis_tree()
        with tree.lock_read(), basis_tree.lock_read():
            changes = tree.iter_changes(basis_tree)
        self.assertTrue(
            only_changes_last_changelog_block(
                tree, tree.basis_tree(), "debian/changelog", list(changes)
            )
        )

    def test_only_new_changelog(self):
        tree = self.make_branch_and_tree(".", format="git")
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/changelog",
                    """\
blah (0.1) unstable; urgency=medium

  * Initial release. (Closes: #911016)

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100
""",
                ),
            ]
        )
        basis_tree = tree.basis_tree()
        with tree.lock_write(), basis_tree.lock_read():
            tree.add(["debian", "debian/changelog"])
            changes = tree.iter_changes(basis_tree)
            self.assertTrue(
                only_changes_last_changelog_block(
                    tree, basis_tree, "debian/changelog", list(changes)
                )
            )

    def test_changes_to_last_only_but_released(self):
        tree = self.make_package_tree()
        self.build_tree_contents(
            [
                (
                    "debian/changelog",
                    """\
blah (0.2) unstable; urgency=medium

  * And a change.

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100

blah (0.1) unstable; urgency=medium

  * Initial release. (Closes: #911016)

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100
""",
                )
            ]
        )
        tree.commit("release")
        self.build_tree_contents(
            [
                (
                    "debian/changelog",
                    """\
blah (0.2) unstable; urgency=medium

  * And a change.
  * Team Upload.

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100

blah (0.1) unstable; urgency=medium

  * Initial release. (Closes: #911016)

 -- Blah <example@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100
""",
                )
            ]
        )
        basis_tree = tree.basis_tree()
        with tree.lock_read(), basis_tree.lock_read():
            changes = tree.iter_changes(basis_tree)
        self.assertFalse(
            only_changes_last_changelog_block(
                tree, tree.basis_tree(), "debian/changelog", list(changes)
            )
        )


class LintianBrushVersion(TestCase):
    def test_matches_package_version(self):
        if not os.path.exists("debian/changelog"):
            self.skipTest(
                "no debian/changelog available. "
                "Running outside of source tree?"
            )
        with open("debian/changelog") as f:
            cl = Changelog(f, max_blocks=1)
        package_version = str(cl.version)
        m = re.match(r'^\d+\.\d+', package_version)
        assert m is not None
        package_version = m.group(0)
        self.assertEqual(package_version, version_string)


class GetCommitterTests(TestCaseWithTransport):
    def test_git_env(self):
        self.overrideEnv("GIT_COMMITTER_NAME", "Some Git Committer")
        self.overrideEnv("GIT_COMMITTER_EMAIL", "committer@example.com")
        tree = self.make_branch_and_tree(".", format="git")
        self.assertEqual(
            "Some Git Committer <committer@example.com>", get_committer(tree)
        )

    def test_git_config(self):
        tree = self.make_branch_and_tree(".", format="git")
        with open(".git/config", "w") as f:
            f.write(
                """\
[user]
   name = Some Other Git Committer
   email = other@example.com
"""
            )
        self.assertEqual(
            "Some Other Git Committer <other@example.com>", get_committer(tree)
        )

    def test_global_stack(self):
        gs = GlobalStack()
        gs.set("email", "Yet Another Committer <yet@example.com>")
        tree = self.make_branch_and_tree(".", format="git")
        self.assertEqual(
            "Yet Another Committer <yet@example.com>", get_committer(tree))


class CertaintySufficientTests(TestCase):
    def test_sufficient(self):
        self.assertTrue(certainty_sufficient("certain", "certain"))
        self.assertTrue(certainty_sufficient("certain", "possible"))
        self.assertTrue(certainty_sufficient("certain", None))
        self.assertTrue(certainty_sufficient("possible", None))
        # TODO(jelmer): Should we really always allow unknown certainties
        # through?
        self.assertTrue(certainty_sufficient(None, "certain"))  # type: ignore

    def test_insufficient(self):
        self.assertFalse(certainty_sufficient("possible", "certain"))


class CertaintyVsConfidenceTests(TestCase):
    def test_certainty_to_confidence(self):
        self.assertEqual(0, certainty_to_confidence("certain"))
        self.assertEqual(1, certainty_to_confidence("confident"))
        self.assertEqual(2, certainty_to_confidence("likely"))
        self.assertEqual(3, certainty_to_confidence("possible"))
        self.assertIs(None, certainty_to_confidence("unknown"))
        self.assertRaises(ValueError, certainty_to_confidence, "blah")


class MinimumCertaintyTests(TestCase):
    def test_minimum(self):
        self.assertEqual("certain", min_certainty([]))
        self.assertEqual("certain", min_certainty(["certain"]))
        self.assertEqual("possible", min_certainty(["possible"]))
        self.assertEqual("possible", min_certainty(["possible", "certain"]))
        self.assertEqual("likely", min_certainty(["likely", "certain"]))
        self.assertEqual(
            "possible", min_certainty(["likely", "certain", "possible"]))


class ParseScriptFixerOutputTests(TestCase):
    def test_simple(self):
        self.assertEqual(
            FixerResult("Do bla", ["tag-a"], "certain"),
            parse_script_fixer_output(
                """\
Do bla
Fixed-Lintian-Tags: tag-a
Certainty: certain
"""
            ),
        )

    def test_unknown_certainty(self):
        self.assertRaises(
            UnsupportedCertainty,
            parse_script_fixer_output,
            """\
Do bla
Fixed-Lintian-Tags: tag-a
Certainty: blah
""",
        )

    def test_default_certainty(self):
        self.assertEqual(
            FixerResult("Do bla", ["tag-a"], None),
            parse_script_fixer_output(
                """\
Do bla
Fixed-Lintian-Tags: tag-a
"""
            ),
        )

    def test_patch_name(self):
        self.assertEqual(
            FixerResult("Do bla", ["tag-a"], "certain", "aname"),
            parse_script_fixer_output(
                """\
Do bla
Fixed-Lintian-Tags: tag-a
Certainty: certain
Patch-Name: aname
"""
            ),
        )


class BaseScriptFixerTests:

    script_fixer_cls: Type[ScriptFixer]

    build_tree_contents: Any
    assertEqual: Callable
    assertIsInstance: Callable
    assertRaises: Callable
    test_dir: str

    def create_fixer(self, code):
        self.build_tree_contents(
            [("script.py", "#!" + sys.executable + "\n" + code)])
        os.chmod("script.py", 0o755)
        fixer = self.script_fixer_cls(
            "fixer", ["a-tag"], os.path.abspath("script.py"))
        self.assertEqual(os.path.abspath("script.py"), fixer.script_path)
        return fixer

    def test_noop(self):
        fixer = self.create_fixer("print('I did not do anything')\n")
        result = fixer.run(self.test_dir, "blah", Version("0.1"), "buster")
        self.assertIsInstance(result, FixerResult)
        self.assertEqual(result.description, "I did not do anything")

    def test_chdir(self):
        fixer = self.create_fixer("import os; print(os.getcwd())\n")
        os.mkdir("subdir")
        os.chdir("subdir")
        result = fixer.run(self.test_dir, "blah", Version("0.1"), "buster")
        self.assertIsInstance(result, FixerResult)
        self.assertEqual(result.description, self.test_dir)

    def test_exception(self):
        fixer = self.create_fixer(
            """\
def foo():
    raise Exception('oops')
foo()
"""
        )
        e = self.assertRaises(
            FixerScriptFailed, fixer.run, self.test_dir, "blah",
            Version("0.1"), "buster")
        self.assertEqual(e.path, fixer.script_path)
        self.assertEqual(e.returncode, 1)
        self.assertEqual(
            e.errors.splitlines()[-2:],
            ["    raise Exception('oops')", "Exception: oops"],
        )


class ScriptFixerTests(
        BaseScriptFixerTests, TestCaseWithTransport):  # type: ignore

    test_dir: str
    script_fixer_cls = ScriptFixer


class PythonScriptFixerTests(
        BaseScriptFixerTests, TestCaseWithTransport):  # type: ignore

    test_dir: str
    script_fixer_cls = PythonScriptFixer   # type: ignore


class SelectFixersTests(TestCase):
    def test_exists(self):
        self.assertEqual(
            ["dummy1"],
            [
                f.name
                for f in select_fixers(
                    [
                        DummyFixer("dummy1", ["some-tag"]),
                        DummyFixer("dummy2", ["other-tag"]),
                    ],
                    names=["dummy1"],
                )
            ],
        )

    def test_missing(self):
        self.assertRaises(
            KeyError, select_fixers, [DummyFixer("dummy", ["some-tag"])],
            names=["other"]
        )

    def test_exclude_missing(self):
        self.assertRaises(
            KeyError,
            select_fixers,
            [DummyFixer("dummy", ["some-tag"])],
            names=["dummy"],
            exclude=["some-other"],
        )

    def test_exclude(self):
        self.assertEqual(
            ["dummy1"],
            [
                f.name
                for f in select_fixers(
                    [
                        DummyFixer("dummy1", ["some-tag"]),
                        DummyFixer("dummy2", ["other-tag"]),
                    ],
                    names=["dummy1", "dummy2"],
                    exclude=["dummy2"],
                )
            ],
        )



#!/usr/bin/python3

import sys
from contextlib import suppress

from debmutate.control import ensure_some_version, get_relation

from lintian_brush.fixer import LintianIssue, control, report_result

try:
    from tomlkit import load
except ModuleNotFoundError:
    sys.exit(2)

try:
    with open('pyproject.toml') as f:
        toml = load(f)
except FileNotFoundError:
    sys.exit(0)


build_backend = toml.get('build-system', {}).get('build-backend')

# See /usr/share/lintian/lib/Lintian/Check/Languages/Python.pm
PREREQUISITE_MAP = {
    'poetry.core.masonry.api': 'python3-poetry-core',
    'flit_core.buildapi': 'flit',
    'setuptools.build_meta': 'python3-setuptools'
}


try:
    prerequisite = PREREQUISITE_MAP[build_backend]
except KeyError:
    sys.exit(2)

with control:
    for field in ['Build-Depends', 'Build-Depends-Indep',
                  'Build-Depends-Arch']:
        with suppress(KeyError):
            if get_relation(control.source.get(field, ''), prerequisite):
                sys.exit(0)
    # TOOD(jelmer): Add file:lineno; requires
    # https://github.com/sdispater/tomlkit/issues/55
    issue = LintianIssue(
        control.source,
        'missing-prerequisite-for-pyproject-backend',
        info=f'{build_backend} (does not satisfy {prerequisite})')
    if issue.should_fix():
        control.source['Build-Depends'] = ensure_some_version(
            control.source.get('Build-Depends', ''), prerequisite)
        issue.report_fixed()

report_result(
    f'Add missing build-dependency on {prerequisite}.\n\n'
    f'This is necessary for build-backend {build_backend} in pyproject.toml')



#!/usr/bin/python3

import sys
from typing import Dict, Set

from lintian_brush.fixer import LintianIssue, control, report_result

removed: Dict[str, Set[str]] = {}

try:
    with control as updater:
        for binary in updater.binaries:
            for field, value in list(binary.items()):
                if updater.source.get(field) == value:
                    del binary[field]
                    removed.setdefault(field, set()).add(binary['Package'])
                    issue = LintianIssue(
                        updater.source,
                        'installable-field-mirrors-source',
                        info='field "{}" in package {}'.format(
                            field, binary['Package']))
                    issue.report_fixed()
except FileNotFoundError:
    sys.exit(0)


if len(removed) == 1:
    (field, binary_packages) = list(removed.items())[0]
    report_result(
        'Remove field {} on binary package{} {} that duplicates source.'
        .format(
            field, 's' if len(binary_packages) != 1 else '',
            ', '.join(sorted(binary_packages))))
elif len(removed) > 1:
    report_result(
        'Remove fields on binary packages that duplicate source.',
        details=['Field {} from {}.'.format(field, ', '.join(sorted(packages)))
                 for (field, packages) in sorted(removed.items())])



#!/usr/bin/python3

import sys
from email.utils import parseaddr

from debmutate.changelog import ChangelogEditor

from lintian_brush.fixer import control, fixed_lintian_tag, report_result

TEAM_UPLOAD_LINE = '  * Team upload.'
uploader_emails = []


with control as updater:
    for entry in updater.source.get('Uploaders', '').split(','):
        uploader_emails.append(parseaddr(entry)[1])


with ChangelogEditor() as updater:
    last_change = updater.changelog[0]
    if last_change.distributions != 'UNRELEASED':
        sys.exit(0)
    has_team_upload = (TEAM_UPLOAD_LINE in last_change.changes())
    name, email = parseaddr(last_change.author)
    if email not in uploader_emails or not has_team_upload:
        sys.exit(0)

    i = last_change._changes.index(TEAM_UPLOAD_LINE)
    del last_change._changes[i]
    if (i > 0 and last_change._changes[i-1] == '' and
            last_change._changes[i] == ''):
        # Also remove the next line, if it's empty
        del last_change._changes[i]
        fixed_lintian_tag('source', 'unnecessary-team-upload')

report_result("Remove unnecessary Team Upload line in changelog.")



#!/usr/bin/python3
import os
import sys

from lintian_brush.fixer import LintianIssue, report_result

if not os.path.exists('debian/pyversions'):
    sys.exit(2)

with open('debian/pyversions') as f:
    pyversions = f.read().strip()

if pyversions.startswith('2.'):
    issue = LintianIssue('source', "debian-pyversions-is-obsolete")
    if issue.should_fix():
        os.unlink('debian/pyversions')
        issue.report_fixed()

report_result("Remove obsolete debian/pyversions.")



#!/usr/bin/python
# Copyright (C) 2022 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for lintian_brush.__main__."""

from breezy.tests import (
    TestCase,
)

from lintian_brush import available_lintian_fixers
from lintian_brush.__main__ import (
    DEFAULT_ADDON_FIXERS,
    DEFAULT_VALUE_LINTIAN_BRUSH,
    DEFAULT_VALUE_LINTIAN_BRUSH_ADDON_ONLY,
    LINTIAN_BRUSH_TAG_DEFAULT_VALUE,
    LINTIAN_BRUSH_TAG_VALUES,
    calculate_value,
    versions_dict,
)


class VersionsDictTests(TestCase):

    def test_expected_types(self):
        d = versions_dict()
        self.assertIsInstance(d, dict, repr(d))
        for k, v in d.items():
            self.assertIsInstance(k, str, repr(d))
            self.assertIsInstance(v, str, repr(d))


class CalculateValueTests(TestCase):

    def test_tags_known(self):
        known_tags = set()
        for fixer in available_lintian_fixers():
            known_tags.update(fixer.lintian_tags)
        for tag in DEFAULT_ADDON_FIXERS:
            self.assertIn(tag, known_tags)
        for tag in LINTIAN_BRUSH_TAG_VALUES:
            self.assertIn(tag, known_tags)

    def test_simple(self):
        self.assertEqual(0, calculate_value(set()))
        self.assertEqual(
            DEFAULT_VALUE_LINTIAN_BRUSH + LINTIAN_BRUSH_TAG_DEFAULT_VALUE,
            calculate_value(['foo']))
        self.assertEqual(
            DEFAULT_VALUE_LINTIAN_BRUSH_ADDON_ONLY
            + LINTIAN_BRUSH_TAG_DEFAULT_VALUE,
            calculate_value([DEFAULT_ADDON_FIXERS[0]]))



#!/usr/bin/python3

from debmutate.control import (
    add_dependency,
    drop_dependency,
    get_relation,
)

from lintian_brush.fixer import LintianIssue, control, report_result

with control as updater:
    try:
        get_relation(
            updater.source.get('Build-Depends-Indep', ''),
            'libmodule-build-perl')
    except KeyError:
        pass
    else:
        issue = LintianIssue(
            updater.source,
            'libmodule-build-perl-needs-to-be-in-build-depends', '')
        if issue.should_fix():
            updater.source['Build-Depends-Indep'] = drop_dependency(
                updater.source['Build-Depends-Indep'],
                'libmodule-build-perl')
            updater.source['Build-Depends'] = add_dependency(
                updater.source.get('Build-Depends', ''),
                'libmodule-build-perl')
            issue.report_fixed()


report_result(
    'Move libmodule-build-perl from Build-Depends-Indep to Build-Depends.')



#!/usr/bin/python3

from typing import Dict, List

from lintian_brush.fixer import (
    LintianIssue,
    linenos_to_ranges,
    report_result,
    shorten_path,
)
from lintian_brush.lintian_overrides import (
    LintianOverride,
    fix_override_info,
    update_overrides,
)

fixed_linenos: Dict[str, List[int]] = {}


def fix_info(path, lineno, override):
    if not override.info:
        return override
    info = fix_override_info(override)
    if info == override.info:
        return override
    issue = LintianIssue(
        (override.type, override.package), 'mismatched-override',
        override.info + ' [%s:%d]' % (path, lineno))
    if issue.should_fix():
        issue.report_fixed()
        fixed_linenos.setdefault(path, []).append(lineno)
        return LintianOverride(
            package=override.package, archlist=override.archlist,
            type=override.type, tag=override.tag,
            info=info)
    return override


update_overrides(fix_info)

if len(fixed_linenos) == 0:
    pass
elif len(fixed_linenos) == 1:
    [(path, linenos)] = fixed_linenos.items()
    report_result(
        "Update lintian override info format in {} on line {}.".format(
            shorten_path(path), ', '.join(linenos_to_ranges(linenos))))
else:
    report_result(
        "Update lintian override info to new format:",
        details=[
            "{}: line {}".format(path, ', '.join(linenos_to_ranges(linenos)))
            for (path, linenos) in fixed_linenos.items()])



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for dealing with control files."""

__all__ = ["guess_repository_url", "determine_browser_url"]

import re
from contextlib import suppress
from typing import Optional
from urllib.parse import urlparse

from ._lintian_brush_rs import guess_repository_url
from .vcs import determine_gitlab_browser_url

TEAM_NAME_MAP = {
    "debian-xml-sgml": "xml-sgml-team",
    "pkg-go": "go-team",
    "pkg-fonts": "fonts-team",
    "pkg-javascript": "js-team",
    "pkg-java": "java-team",
    "pkg-mpd": "mpd-team",
    "pkg-electronics": "electronics-team",
    "pkg-xfce": "xfce-team",
    "pkg-lxc": "lxc-team",
    "debian-science": "science-team",
    "pkg-games": "games-team",
    "pkg-bluetooth": "bluetooth-team",
    "debichem": "debichem-team",
    "openstack": "openstack-team",
    "pkg-kde": "qt-kde-team",
    "debian-islamic": "islamic-team",
    "pkg-lua": "lua-team",
    "pkg-xorg": "xorg-team",
    "debian-astro": "debian-astro-team",
    "pkg-icecast": "multimedia-team",
    "glibc-bsd": "bsd-team",
    "pkg-nvidia": "nvidia-team",
    "pkg-llvm": "llvm-team",
    "pkg-nagios": "nagios-team",
    "pkg-sugar": "pkg-sugar-team",
    "pkg-phototools": "debian-phototools-team",
    "pkg-netmeasure": "ineteng-team",
    "pkg-hamradio": "debian-hamradio-team",
    "pkg-sass": "sass-team",
    "pkg-rpm": "pkg-rpm-team",
    "tts": "tts-team",
    "python-apps": "python-team/applications",
    "pkg-monitoring": "monitoring-team",
    "pkg-perl": "perl-team/modules",
    "debian-iot": "debian-iot-team",
    "pkg-bitcoin": "cryptocoin-team",
    "pkg-cyrus-imapd": "debian",
    "pkg-dns": "dns-team",
    "pkg-freeipa": "freeipa-team",
    "pkg-ocaml-team": "ocaml-team",
    "pkg-vdr-dvb": "vdr-team",
    "debian-in": "debian-in-team",
    "pkg-octave": "pkg-octave-team",
    "pkg-postgresql": "postgresql",
    "pkg-grass": "debian-gis-team",
    "pkg-evolution": "gnome-team",
    "pkg-gnome": "gnome-team",
    "pkg-exppsy": "neurodebian-team",
    "pkg-voip": "pkg-voip-team",
    "pkg-privacy": "pkg-privacy-team",
    "pkg-libvirt": "libvirt-team",
    "debian-ha": "ha-team",
    "debian-lego": "debian-lego-team",
    "calendarserver": "calendarserver-team",
    "3dprinter": "3dprinting-team",
    "pkg-multimedia": "multimedia-team",
    "pkg-emacsen": "emacsen-team",
    "pkg-haskell": "haskell-team",
    "pkg-gnutls": "gnutls-team",
    "pkg-mysql": "mariadb-team",
    "pkg-php": "php-team",
    "pkg-qemu": "qemu-team",
    "pkg-xmpp": "xmpp-team",
    "uefi": "efi-team",
    "pkg-manpages-fr": "l10n-fr-team",
    "pkg-proftpd": "debian-proftpd-team",
    "pkg-apache": "apache-team",
}

GIT_PATH_RENAMES = {
    "pkg-kde/applications": "qt-kde-team/kde",
    "3dprinter/packages": "3dprinting-team",
    "pkg-emacsen/pkg": "emacsen-team",
    "debian-astro/packages": "debian-astro-team",
    "debian-islamic/packages": "islamic-team",
    "debichem/packages": "debichem-team",
    "pkg-privacy/packages": "pkg-privacy-team",
    "pkg-cli-libs/packages": "dotnet-team",
}


def determine_browser_url(url: str) -> str:
    """Determine the browser URL from a regular Git URL.

    Args:
      url: Regular URL
    Returns:
      a browser URL
    """
    return determine_gitlab_browser_url(url)


def _salsa_path_from_alioth_url(  # noqa: C901
    vcs_type: Optional[str], alioth_url: str
) -> Optional[str]:
    if vcs_type is None:
        return None
    # These two regular expressions come from vcswatch:
    # https://salsa.debian.org/qa/qa/blob/master/data/vcswatch/vcswatch#L165
    if vcs_type.lower() == "git":
        pat = ("(https?|git)://(anonscm|git).debian.org/"
               "(cgit/|git/)?collab-maint/")
        if re.match(pat, alioth_url):
            return re.sub(pat, "debian/", alioth_url)
        pat = "(https?|git)://(anonscm|git).debian.org/(cgit/|git/)?users/"
        if re.match(pat, alioth_url):
            return re.sub(pat, "", alioth_url)
        m = re.match(
            "(https?|git)://(anonscm|git).debian.org/(cgit/|git/)?(.+)",
            alioth_url
        )
        if m:
            parts = m.group(4).split("/")
            for i in range(len(parts), 0, -1):
                subpath = "/".join(parts[:i])
                with suppress(KeyError):
                    return (GIT_PATH_RENAMES[subpath]
                            + "/" + "/".join(parts[i:]))
        m = re.match(
            "(https?|git)://(anonscm|git).debian.org/(cgit/|git/)?([^/]+)/",
            alioth_url
        )
        if m and m.group(4) == "debian-in" and "fonts-" in alioth_url:
            return re.sub(m.re, "fonts-team/", alioth_url)
        if m and m.group(4) in TEAM_NAME_MAP:
            new_name = TEAM_NAME_MAP[m.group(4)]
            return re.sub(m.re, new_name + "/", alioth_url)
        m = re.match(
            "https?://alioth.debian.org/anonscm/(git/|cgit/)?([^/]+)/",
            alioth_url
        )
        if m and m.group(2) in TEAM_NAME_MAP:
            new_name = TEAM_NAME_MAP[m.group(2)]
            return re.sub(m.re, new_name + "/", alioth_url)

    if vcs_type.lower() == "svn":
        if alioth_url.startswith("svn://svn.debian.org/pkg-perl/trunk"):
            return alioth_url.replace(
                "svn://svn.debian.org/pkg-perl/trunk",
                "perl-team/modules/packages"
            )
        if alioth_url.startswith("svn://svn.debian.org/pkg-lua/packages"):
            return alioth_url.replace(
                "svn://svn.debian.org/pkg-lua/packages", "lua-team"
            )
        parsed_url = urlparse(alioth_url)
        if parsed_url.scheme == "svn" and parsed_url.netloc in (
            ("svn.debian.org", "anonscm.debian.org")
        ):
            parts = parsed_url.path.strip("/").split("/")
            if parts[0] == "svn":
                parts.pop(0)
            if (len(parts) == 3 and parts[0] in TEAM_NAME_MAP
                    and parts[2] == "trunk"):
                return f"{TEAM_NAME_MAP[parts[0]]}/{parts[1]}"
            if (len(parts) == 3 and parts[0] in TEAM_NAME_MAP
                    and parts[1] == "trunk"):
                return f"{TEAM_NAME_MAP[parts[0]]}/{parts[2]}"
            if (
                len(parts) == 4
                and parts[0] in TEAM_NAME_MAP
                and parts[1] == "packages"
                and parts[3] == "trunk"
            ):
                return f"{TEAM_NAME_MAP[parts[0]]}/{parts[2]}"
            if (
                len(parts) == 4
                and parts[0] in TEAM_NAME_MAP
                and parts[1] == "trunk"
                and parts[2] == "packages"
            ):
                return f"{TEAM_NAME_MAP[parts[0]]}/{parts[3]}"
            if (len(parts) > 3 and parts[0] in TEAM_NAME_MAP
                    and parts[-2] == "trunk"):
                return f"{TEAM_NAME_MAP[parts[0]]}/{parts[-1]}"
            if (
                len(parts) == 3
                and parts[0] in TEAM_NAME_MAP
                and parts[1] in ("packages", "unstable")
            ):
                return f"{TEAM_NAME_MAP[parts[0]]}/{parts[2]}"
    return None


def salsa_url_from_alioth_url(vcs_type: str, alioth_url: str) -> Optional[str]:
    """Guess the salsa URL from an alioth URL.

    Args:
      vcs_type: VCS type
      alioth_url: Alioth URL
    Returns:
      Salsa URL
    """
    path = _salsa_path_from_alioth_url(vcs_type, alioth_url)
    if path is None:
        return None
    path = path.strip("/")
    if not path.endswith(".git"):
        path = path + ".git"
    return "https://salsa.debian.org/" + path



#!/usr/bin/python3

import sys

from debmutate.control import drop_dependency

from lintian_brush.fixer import LintianIssue, control, report_result

try:
    with open('debian/rules', 'rb') as f:
        for line in f:
            if b'/usr/share/cdbs/' in line:
                uses_cdbs = True
                break
        else:
            uses_cdbs = False
except FileNotFoundError:
    # Unsure whether it actually needs cdbs
    sys.exit(2)

if not uses_cdbs:
    with control as updater:
        for field in ["Build-Depends", "Build-Depends-Indep"]:
            new_depends = drop_dependency(
                updater.source.get(field, ""), "cdbs")
            if new_depends != updater.source.get(field, ""):
                issue = LintianIssue(
                    updater.source, 'unused-build-dependency-on-cdbs', '')
                if issue.should_fix():
                    updater.source[field] = new_depends
                    if not updater.source[field]:
                        del updater.source[field]
                    issue.report_fixed()

report_result("Drop unused build-dependency on cdbs.")



#!/usr/bin/python3

import os
import shlex
import sys

from debmutate._rules import Makefile, Rule, dh_invoke_get_with
from debmutate.control import (
    PkgRelation,
    add_dependency,
    is_relation_implied,
    parse_relations,
)

from lintian_brush.fixer import LintianIssue, control, report_result
from lintian_brush.lintian import LINTIAN_DATA_PATH, dh_addons, dh_commands

COMMAND_TO_DEP = {}


if not os.path.isdir(LINTIAN_DATA_PATH):
    # lintian doesn't appear to be installed
    sys.exit(2)


for command, info in dh_commands().items():
    COMMAND_TO_DEP[command] = ' | '.join(info['installed_by'])

# Copied from /usr/share/lintian/lib/Lintian/Check/Debhelper.pm
COMMAND_TO_DEP.update({
    'dh_apache2': 'dh-apache2 | apache2-dev',
    'dh_autoreconf_clean':
        'dh-autoreconf | debhelper (>= 9.20160403~) | debhelper-compat',
    'dh_autoreconf':
        'dh-autoreconf | debhelper (>= 9.20160403~) | debhelper-compat',
    'dh_dkms': 'dkms | dh-sequence-dkms',
    'dh_girepository': 'gobject-introspection | dh-sequence-gir',
    'dh_gnome': 'gnome-pkg-tools | dh-sequence-gnome',
    'dh_gnome_clean': 'gnome-pkg-tools | dh-sequence-gnome',
    'dh_lv2config': 'lv2core',
    'dh_make_pgxs': 'postgresql-server-dev-all | postgresql-all',
    'dh_nativejava': 'gcj-native-helper | default-jdk-builddep',
    'dh_pgxs_test': 'postgresql-server-dev-all | postgresql-all',
    'dh_python2': 'dh-python | dh-sequence-python2',
    'dh_python3': 'dh-python | dh-sequence-python3',
    'dh_sphinxdoc': 'sphinx | python-sphinx | python3-sphinx',
    'dh_xine': 'libxine-dev | libxine2-dev',
})


ADDON_TO_DEP = {}
for addon, info in dh_addons().items():
    ADDON_TO_DEP[addon] = ' | '.join(info['installed_by'])


ADDON_TO_DEP.update({
    # Copied from /usr/share/lintian/lib/Lintian/Check/Debhelper.pm
    'ada_library': 'dh-ada-library | dh-sequence-ada-library',
    'apache2': 'dh-apache2 | apache2-dev',
    'autoreconf':
        'dh-autoreconf | debhelper (>= 9.20160403~) | debhelper-compat',
    'cli': 'cli-common-dev | dh-sequence-cli',
    'dwz': 'debhelper | debhelper-compat | dh-sequence-dwz',
    'installinitramfs':
        'debhelper | debhelper-compat | dh-sequence-installinitramfs',
    'gnome': 'gnome-pkg-tools | dh-sequence-gnome',
    'lv2config': 'lv2core',
    'nodejs': 'pkg-js-tools | dh-sequence-nodejs',
    'perl_dbi': 'libdbi-perl | dh-sequence-perl-dbi',
    'perl_imager': 'libimager-perl | dh-sequence-perl-imager',
    'pgxs': 'postgresql-server-dev-all | postgresql-all',
    'pgxs_loop':  'postgresql-server-dev-all | postgresql-all',
    'pypy': 'dh-python | dh-sequence-pypy',
    'python2': 'python2:any | python2-dev:any | dh-sequence-python2',
    'python3':
        'python3:any | python3-all:any | python3-dev:any | '
        'python3-all-dev:any | dh-sequence-python3',
    'scour':  'scour | python-scour | dh-sequence-scour',
    'sphinxdoc':
        'sphinx | python-sphinx | python3-sphinx | dh-sequence-sphinxdoc',
    'systemd':
        'debhelper (>= 9.20160709~) | debhelper-compat | '
        'dh-sequence-systemd | dh-systemd',
    'vim_addon': 'dh_vim-addon | dh-sequence-vim-addon',
})

need = []

try:
    mf = Makefile.from_path('debian/rules')
except FileNotFoundError:
    sys.exit(0)

for entry in mf.contents:
    if isinstance(entry, Rule):
        for command in entry.commands():
            if command.lstrip().startswith(b'#'):
                continue
            try:
                executable = shlex.split(command.decode('utf-8', 'replace'))[0]
            except (ValueError, IndexError):
                continue
            try:
                dep = COMMAND_TO_DEP[executable]
            except KeyError:
                pass
            else:
                issue = LintianIssue(
                    'source', 'missing-build-dependency-for-dh_-command',
                    f'{executable} => {dep}')
                if not issue.should_fix():
                    continue
                need.append((dep, [issue], 'command', executable))
            if executable == 'dh' or executable.startswith('dh_'):
                for addon in dh_invoke_get_with(command):
                    try:
                        dep = ADDON_TO_DEP[addon]
                    except KeyError:
                        pass
                    else:
                        issue = LintianIssue(
                            'source', 'missing-build-dependency-for-dh-addon',
                            f'{addon} => {dep}')
                        need.append((dep, [issue], 'addon', addon))


if not need:
    sys.exit(0)

changed = []

with control as updater:
    for deps, issues, kind, name in need:
        parsed = PkgRelation.parse(deps)
        is_implied = False

        if is_relation_implied(parsed, 'debhelper'):
            is_implied = True

        for field in ['Build-Depends', 'Build-Depends-Indep',
                      'Build-Depends-Arch']:
            for _unused1, existing, _unused2 in parse_relations(
                    updater.source.get(field, '')):
                if is_relation_implied(parsed, existing):
                    is_implied = True

        if not is_implied:
            build_deps = updater.source.get('Build-Depends', '')
            updater.source['Build-Depends'] = add_dependency(build_deps, deps)
            for issue in issues:
                issue.report_fixed()
            changed.append((deps, issue, kind, name))

if len(changed) == 1:
    (dep, issue, kind, name) = changed[0]
    report_result(
        f'Add missing build dependency on {dep} for {kind} {name}.')
else:
    report_result(
        'Add missing build dependencies:' +
        '\n'.join(f'* {dep} for {kind} {name}'
                  for (dep, issue, kind, name) in changed))



#!/usr/bin/python3

from contextlib import suppress

from debmutate.copyright import CopyrightEditor, NotMachineReadableError

from debian.copyright import License
from lintian_brush.fixer import fixed_lintian_tag, report_result
from lintian_brush.licenses import load_spdx_data

RENAMES = {k.lower(): v for k, v in {
  'Creative Commons Attribution Share-Alike (CC-BY-SA) v3.0': 'CC-BY-SA-3.0',
  'Apache License Version 2.0': 'Apache-2.0',
}.items()}


# TODO(jelmer): Ideally we'd get the list of standard SPDX
REPLACE_SPACES = {
  'public-domain',
  'mit-style',
  'bsd-style',
}

spdx_data = load_spdx_data()

RENAMES.update(
    {license['name'].lower(): license_id
     for license_id, license in spdx_data['licenses'].items()})
REPLACE_SPACES.update(set(spdx_data['licenses']))
REPLACE_SPACES = {license_id.lower() for license_id in REPLACE_SPACES}
for license_id in list(REPLACE_SPACES):
    if license_id.endswith('.0'):
        REPLACE_SPACES.add(license_id[:-2])


def fix_spaces(copyright):
    for paragraph in copyright.all_paragraphs():
        if not paragraph.license:
            continue
        if ' ' not in paragraph.license.synopsis:
            continue
        ors = paragraph.license.synopsis.replace(' | ', ' or ').split(' or ')
        names = []
        for name in ors:
            if name.lower() in RENAMES:
                name = RENAMES[name.lower()]
            elif name.replace(' ', '-').lower() in REPLACE_SPACES:
                name = name.replace(' ', '-')
            names.append(name)
        newsynopsis = ' or '.join(names)
        if newsynopsis != paragraph.license.synopsis:
            fixed_lintian_tag(
                'source', 'space-in-std-shortname-in-dep5-copyright',
                '%s (line XX)' % paragraph.license.synopsis)
            paragraph.license = License(newsynopsis, paragraph.license.text)


with suppress(FileNotFoundError, NotMachineReadableError), \
        CopyrightEditor() as updater:
    fix_spaces(updater.copyright)

report_result(
    'Replace spaces in short license names with dashes.')



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for lintian_brush.patches."""

from io import BytesIO

from breezy.patches import (
    Patch,
    parse_patch,
)
from breezy.tests import (
    TestCase,
    TestCaseWithTransport,
)

from ..patches import (
    AppliedPatches,
    find_common_patch_suffix,
    find_patch_base,
    find_patches_branch,
    read_quilt_patches,
    read_quilt_series,
    tree_non_patches_changes,
    tree_patches_directory,
    upstream_with_applied_patches,
)


class ReadSeriesFileTests(TestCase):
    def test_comment(self):
        f = BytesIO(
            b"""\
# This file intentionally left blank.
"""
        )
        self.assertEqual(
            [("This", True, ["file", "intentionally", "left", "blank."])],
            list(read_quilt_series(f)),
        )

    def test_empty(self):
        f = BytesIO(b"\n")
        self.assertEqual([], list(read_quilt_series(f)))

    def test_empty_line(self):
        f = BytesIO(
            b"""\
patch1

patch2
"""
        )
        self.assertEqual(
            [
                ("patch1", False, []),
                ("patch2", False, []),
            ],
            list(read_quilt_series(f)),
        )

    def test_options(self):
        f = BytesIO(
            b"""\
name -p0
name2
"""
        )
        self.assertEqual(
            [
                ("name", False, ["-p0"]),
                ("name2", False, []),
            ],
            list(read_quilt_series(f)),
        )


class FindPatchBaseTests(TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        self.tree = self.make_branch_and_tree(".")
        self.upstream_revid = self.tree.commit("upstream")
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/changelog",
                    """\
blah (0.38) unstable; urgency=medium

  * Fix something

 -- Jelmer Vernooij <jelmer@debian.org>  Sat, 19 Oct 2019 15:21:53 +0000
 """,
                ),
            ]
        )
        self.tree.add(["debian", "debian/changelog"])

    def test_none(self):
        self.assertEqual(None, find_patch_base(self.tree))

    def test_upstream_dash(self):
        self.tree.branch.tags.set_tag("upstream-0.38", self.upstream_revid)
        self.assertEqual(self.upstream_revid, find_patch_base(self.tree))


class FindPatchBranchTests(TestCaseWithTransport):
    def make_named_branch_and_tree(self, path, name):
        dir = self.make_controldir(path)
        dir.create_repository()
        branch = dir.create_branch(name=name)
        dir.set_branch_reference(branch)
        return dir.create_workingtree()

    def test_none(self):
        tree = self.make_branch_and_tree(".")
        self.assertIs(None, find_patches_branch(tree))

    def test_patch_queue(self):
        master = self.make_named_branch_and_tree(".", name="master")
        master.branch.controldir.create_branch(name="patch-queue/master")

        self.assertEqual(
            "patch-queue/master", find_patches_branch(master).name)

    def test_patched_master(self):
        master = self.make_named_branch_and_tree(".", name="master")
        master.branch.controldir.create_branch(name="patched")
        self.assertEqual("patched", find_patches_branch(master).name)

    def test_patched_other(self):
        other = self.make_named_branch_and_tree(".", name="other")
        other.branch.controldir.create_branch(name="patched-other")
        self.assertEqual("patched-other", find_patches_branch(other).name)


class AppliedPatchesTests(TestCaseWithTransport):
    def test_apply_simple(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("a", "a\n")])
        tree.add("a")
        tree.commit("Add a")
        patch = parse_patch(
            b"""\
--- a/a
+++ b/a
@@ -1 +1 @@
-a
+b
""".splitlines(
                True
            )
        )
        with AppliedPatches(tree, [patch]) as newtree:
            self.assertEqual(b"b\n", newtree.get_file_text("a"))

    def test_apply_delete(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("a", "a\n")])
        tree.add("a")
        tree.commit("Add a")
        patch = parse_patch(
            b"""\
--- a/a
+++ /dev/null
@@ -1 +0,0 @@
-a
""".splitlines(
                True
            )
        )
        with AppliedPatches(tree, [patch]) as newtree:
            self.assertFalse(newtree.has_filename("a"))

    def test_apply_add(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("a", "a\n")])
        tree.add("a")
        tree.commit("Add a")
        patch = parse_patch(
            b"""\
--- /dev/null
+++ b/b
@@ -0,0 +1 @@
+b
""".splitlines(
                True
            )
        )
        with AppliedPatches(tree, [patch]) as newtree:
            self.assertEqual(b"b\n", newtree.get_file_text("b"))


class ReadQuiltPatchesTests(TestCaseWithTransport):
    def test_read_patches(self):
        patch = """\
--- a/a
+++ b/a
@@ -1,5 +1,5 @@
 line 1
 line 2
-line 3
+new line 3
 line 4
 line 5
"""
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                ("debian/",),
                ("debian/patches/",),
                ("debian/patches/series", "foo\n"),
                ("debian/patches/foo", patch),
            ]
        )
        tree.add(
            ["debian", "debian/patches", "debian/patches/series",
             "debian/patches/foo"]
        )
        tree.commit("add patch")
        patches = list(read_quilt_patches(tree))
        self.assertEqual(1, len(patches))
        self.assertIsInstance(patches[0], Patch)
        self.assertEqual(patch.encode("utf-8"), patches[0].as_bytes())

    def test_no_series_file(self):
        t = self.make_branch_and_tree(".")
        self.assertEqual([], list(read_quilt_patches(t)))

    def test_comments(self):
        t = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                ("debian/",),
                ("debian/patches/",),
                ("debian/patches/series",
                 "# This file intentionally left blank.\n"),
            ]
        )
        self.assertEqual([], list(read_quilt_patches(t)))


class UpstreamWithAppliedPatchesTests(TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        self.tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("afile", "some line\n")])
        self.tree.add("afile")
        self.upstream_revid = self.tree.commit("upstream")
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/changelog",
                    """\
blah (0.38) unstable; urgency=medium

  * Fix something

 -- Jelmer Vernooij <jelmer@debian.org>  Sat, 19 Oct 2019 15:21:53 +0000
 """,
                ),
                ("debian/patches/",),
                ("debian/patches/series", "1.patch\n"),
                (
                    "debian/patches/1.patch",
                    """\
--- a/afile
+++ b/afile
@@ -1 +1 @@
-some line
+another line
--- /dev/null
+++ b/newfile
@@ -0,0 +1 @@
+new line
""",
                ),
                ("unchangedfile", "unchanged\n"),
            ]
        )
        self.tree.add(["debian", "debian/changelog", "unchangedfile"])

    def test_upstream_branch(self):
        self.tree.branch.tags.set_tag("upstream/0.38", self.upstream_revid)
        patches = list(read_quilt_patches(self.tree))
        with upstream_with_applied_patches(self.tree, patches) as t:
            self.assertEqual(b"another line\n", t.get_file_text("afile"))
            self.assertEqual(b"new line\n", t.get_file_text("newfile"))
            # TODO(jelmer): PreviewTree appears to be broken
            # self.assertEqual(b'unchanged\n',
            #                  t.get_file_text('unchangedfile'))


class TreePatchesNonPatchesTests(TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        self.tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("afile", "some line\n")])
        self.tree.add("afile")
        self.upstream_revid = self.tree.commit("upstream")
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/changelog",
                    """\
blah (0.38) unstable; urgency=medium

  * Fix something

 -- Jelmer Vernooij <jelmer@debian.org>  Sat, 19 Oct 2019 15:21:53 +0000
 """,
                ),
                ("debian/patches/",),
                ("debian/patches/series", "1.patch\n"),
                (
                    "debian/patches/1.patch",
                    """\
--- a/afile
+++ b/afile
@@ -1 +1 @@
-some line
+another line
""",
                ),
            ]
        )
        self.tree.add(["debian", "debian/changelog"])

    def test_no_delta(self):
        self.tree.branch.tags.set_tag("upstream/0.38", self.upstream_revid)
        self.assertEqual(
            [], list(tree_non_patches_changes(self.tree, "debian/patches"))
        )

    def test_delta(self):
        self.tree.branch.tags.set_tag("upstream/0.38", self.upstream_revid)
        self.build_tree_contents([("anotherfile", "blah")])
        self.tree.add("anotherfile")
        self.assertEqual(
            1, len(list(tree_non_patches_changes(self.tree, "debian/patches")))
        )


class FindCommonPatchSuffixTests(TestCase):
    def test_simple(self):
        self.assertEqual(
            ".blah", find_common_patch_suffix(["foo.blah", "series"]))
        self.assertEqual(
            ".patch", find_common_patch_suffix(["foo.patch", "series"]))
        self.assertEqual(
            ".patch", find_common_patch_suffix(["series"]))
        self.assertEqual(
            ".blah", find_common_patch_suffix(["series"], ".blah"))
        self.assertEqual(
            "", find_common_patch_suffix(["series", "foo", "bar"]))
        self.assertEqual(
            ".patch",
            find_common_patch_suffix(["series", "foo.patch", "bar.patch"])
        )


class TreePatchesDirectoryTests(TestCaseWithTransport):
    def test_none(self):
        tree = self.make_branch_and_tree(".")
        self.assertEqual("debian/patches", tree_patches_directory(tree))

    def test_default(self):
        tree = self.make_branch_and_tree(".")
        tree.mkdir("debian")
        tree.mkdir("debian/patches")
        self.assertEqual("debian/patches", tree_patches_directory(tree))

    def test_custom(self):
        tree = self.make_branch_and_tree(".")
        tree.mkdir("debian")
        tree.mkdir("debian/patches-applied")
        self.build_tree_contents(
            [
                (
                    "debian/rules",
                    """\

export QUILT_PATCH_DIR := debian/patches-applied

all:

blah: bloe
\tfoo
""",
                )
            ]
        )
        tree.add("debian/rules")
        self.assertEqual(
            "debian/patches-applied", tree_patches_directory(tree))



#!/usr/bin/python3

import os
import sys

from lintian_brush import (
    DEFAULT_MINIMUM_CERTAINTY,
    min_certainty,
)
from lintian_brush.multiarch_hints import (
    apply_multiarch_hints,
    download_multiarch_hints,
    multiarch_hints_by_binary,
    parse_multiarch_hints,
)

if os.environ.get('NET_ACCESS', 'disallow') != 'allow':
    sys.exit(0)

with download_multiarch_hints() as f:
    hints = multiarch_hints_by_binary(parse_multiarch_hints(f))

certainty = os.environ.get('MINIMUM_CERTAINTY', DEFAULT_MINIMUM_CERTAINTY)

changes = apply_multiarch_hints(hints, certainty)

overall_certainty = min_certainty(
    [certainty for (binary, hint, description, certainty) in changes])

print("Apply multi-arch hints.")
print("")

for (binary, _hint, description, _certainty) in changes:
    print("* {}: {}".format(binary['Package'], description))



#!/usr/bin/python3

import os
import subprocess
import sys

from debmutate.watch import parse_watch_file

from lintian_brush.fixer import (
    LintianIssue,
    net_access_allowed,
    report_result,
)

if not os.path.exists('debian/watch'):
    sys.exit(0)


def watchfile_has_http():
    with open('debian/watch') as f:
        wf = parse_watch_file(f)

    if not wf:
        return False

    for entry in wf:
        if entry.url.startswith('http://'):
            return True
    else:
        # No plain HTTP URLs
        return False


def update_watchfile(fn):
    with open('debian/watch') as f:
        old = f.readlines()

    new = []
    for line in old:
        try:
            (bef, aft) = line.split('#', 1)
        except ValueError:
            bef = line
            aft = None
        newbef = fn(bef)
        if newbef != bef:
            issue = LintianIssue(
                'source', 'debian-watch-uses-insecure-uri',
                bef)
            if issue.should_fix():
                if aft is not None:
                    new.append('#'.join([newbef, aft]))
                else:
                    new.append(newbef)
                issue.report_fixed()
            else:
                new.append(line)
        else:
            new.append(line)

    if old != new:
        with open('debian/watch', 'w') as f:
            f.writelines(new)
        return True
    return False


if not watchfile_has_http():
    sys.exit(0)


# We hardcode the replacements for some sites, since these testsuite uses
# these. The method below (involving uscan) doesn't work from e.g. sbuild
# hosts.
def stock_replace(line):
    for hostname in ['code.launchpad.net', 'launchpad.net', 'ftp.gnu.org']:
        line = line.replace('http://%s/' % hostname, 'https://%s/' % hostname)
    return line


update_watchfile(stock_replace)


report_result("Use secure URI in debian/watch.")


if not watchfile_has_http():
    sys.exit(0)


if not net_access_allowed():
    sys.exit(0)


def run_uscan_dehs():
    return subprocess.check_output(
        ['uscan', '--dehs', '--report'], stderr=subprocess.PIPE)


try:
    before = run_uscan_dehs()
except subprocess.CalledProcessError:
    # Before doesn't work :(
    sys.exit(0)


def replace_all(line):
    return line.replace('http://', 'https://')


if not update_watchfile(replace_all):
    sys.exit(0)

try:
    after = run_uscan_dehs()
except subprocess.CalledProcessError:
    sys.exit(2)

# uscan creates backup files.
for path in [
        'debian/upstream/signing-key.pgp.backup',
        'debian/upstream-signing-key.pgp.backup']:
    if os.path.exists(path):
        os.unlink(path)

# Make sure that reports are same up to http/https substitution in URL.
if before.replace(b'http://', b'https://') != after:
    # Couldn't do anything :(
    sys.exit(2)



#!/usr/bin/python3

import sys

from debmutate.deb822 import Deb822Editor

from lintian_brush.fixer import report_result, vendor
from lintian_brush.lintian import known_tests_control_fields

try:
    from Levenshtein import distance
except ImportError:
    sys.exit(2)


# See https://people.debian.org/~mpitt/autopkgtest/README.package-tests.html
valid_field_names = set(known_tests_control_fields(vendor()))

case_fixed = set()
typo_fixed = set()


try:
    with Deb822Editor('debian/tests/control') as updater:
        for paragraph in updater.paragraphs:
            for field in paragraph:
                if field in valid_field_names:
                    continue
                for option in valid_field_names:
                    if distance(field, option) == 1:
                        value = paragraph[field]
                        del paragraph[field]
                        paragraph[option] = value
                        if option.lower() == field.lower():
                            case_fixed.add((field, option))
                        else:
                            typo_fixed.add((field, option))
                        break
except FileNotFoundError:
    sys.exit(0)

if case_fixed:
    kind = 'case' + ('s' if len(case_fixed) > 1 else '')
else:
    kind = ''
if typo_fixed:
    if case_fixed:
        kind += ' and '
    kind += 'typo' + ('s' if len(typo_fixed) > 1 else '')

fixed_str = ', '.join(
    [f'{old} ⇒ {new}'
     for (old, new) in sorted(list(case_fixed) + list(typo_fixed))])

report_result(
    f'Fix field name {kind} in debian/tests/control ({fixed_str}).')



#!/usr/bin/python3

# Copyright (C) 2020 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility for dropping unnecessary constraints."""

import json
import logging
import os
import sys
from contextlib import suppress
from typing import Callable, Dict, List, Optional, Tuple, Union

from breezy.commit import NullCommitReporter, PointlessCommit
from breezy.workingtree import WorkingTree
from debmutate.control import (
    ControlEditor,
    PkgRelation,
    format_relations,
    guess_template_type,
    parse_relations,
)
from debmutate.deb822 import ChangeConflict
from debmutate.debhelper import MaintscriptEditor
from debmutate.reformatting import (
    FormattingUnpreservable,
    GeneratedFile,
    check_generated_file,
)

from debian.deb822 import Deb822Dict
from debian.debian_support import Version

from . import (
    NotDebianPackage,
    control_file_present,
    control_files_in_root,
    get_committer,
    is_debcargo_package,
)
from .changelog import add_changelog_entry
from .debhelper import drop_obsolete_maintscript_entries
from .svp import (
    report_fatal,
    svp_enabled,
)

DEFAULT_VALUE_MULTIARCH_HINT = 30


class UddError(Exception):
    """Error accessing UDD."""


_changelog_policy_noted = False


def _note_changelog_policy(policy, msg):
    global _changelog_policy_noted
    if not _changelog_policy_noted:
        if policy:
            extra = "Specify --no-update-changelog to override."
        else:
            extra = "Specify --update-changelog to override."
        logging.info("%s %s", msg, extra)
    _changelog_policy_noted = True


class Action:
    """Action."""

    def __init__(self, rel):
        self.rel = rel

    def __repr__(self):
        return f"<{type(self).__name__}({PkgRelation.str(self.rel)!r})>"

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.rel == other.rel

    def json(self):
        raise NotImplementedError(self.json)


class DropEssential(Action):
    """Drop dependency on essential package."""

    def __str__(self):
        return "Drop dependency on essential package %s" % (
            PkgRelation.str(self.rel))

    def json(self):
        return ("drop-essential", PkgRelation.str(self.rel))


class DropMinimumVersion(Action):
    """Drop minimumversion."""

    def __str__(self):
        return "Drop versioned constraint on %s" % PkgRelation.str(self.rel)

    def json(self):
        return ("drop-minimum-version", PkgRelation.str(self.rel))


class DropTransition(Action):
    """Drop dependency on dummy transitional package."""

    def __str__(self):
        return (
            "Drop dependency on transitional package %s" %
            PkgRelation.str(self.rel))

    def json(self):
        return ("drop-transitional", PkgRelation.str(self.rel))


class ReplaceTransition(Action):
    """Replace dependency on dummy transitional package."""

    def __init__(self, rel, replacement):
        super().__init__(rel)
        self.replacement = replacement

    def __str__(self):
        return (
            "Replace dependency on transitional package {} "
            "with replacement {}".format(
                PkgRelation.str(self.rel), name_list([
                    PkgRelation.str(p) for p in self.replacement])))

    def json(self):
        return ("inline-transitional", PkgRelation.str(self.rel),
                [PkgRelation.str(p) for p in self.replacement])

    def __repr__(self):
        return "<{}({!r}, {!r})>".format(
            type(self).__name__, PkgRelation.str(self.rel),
            [PkgRelation(p) for p in self.replacement])

    def __eq__(self, other):
        return (
            isinstance(other, type(self)) and self.rel == other.rel
            and self.replacement == other.replacement)


class DropObsoleteConflict(Action):
    """Drop conflict with obsolete package."""

    def __str__(self):
        return "Drop conflict with removed package %s" % (
            PkgRelation.str(self.rel))

    def json(self):
        return ("drop-obsolete-conflict", PkgRelation.str(self.rel))


def depends_obsolete(
        latest_version: Version, kind: str, req_version: Version) -> bool:
    req_version = Version(req_version)
    if kind == ">=":  # noqa: SIM116
        return latest_version >= req_version
    elif kind == ">>":
        return latest_version > req_version
    elif kind == "=":
        return False
    return False


def conflict_obsolete(
        latest_version: Version, kind: str, req_version: Version):
    req_version = Version(req_version)
    if kind == "<<":
        return latest_version >= req_version
    elif kind in ("<=", "="):
        return latest_version > req_version
    return False


def _package_version(package: str, release: str) -> Optional[Version]:
    from .udd import connect_udd_mirror

    conn = connect_udd_mirror()
    with conn.cursor() as cursor:
        cursor.execute(
            "SELECT version FROM packages "
            "WHERE package = %s AND release = %s",
            (package, release))
        row = cursor.fetchone()
        if row is not None:
            return Version(row[0])
        return None


def _package_provides(
        package: str, release: str) -> Optional[List[PkgRelation]]:
    from .udd import connect_udd_mirror

    conn = connect_udd_mirror()
    with conn.cursor() as cursor:
        cursor.execute(
            "SELECT provides FROM packages "
            "WHERE package = %s AND release = %s",
            (package, release))
        row = cursor.fetchone()
        if row is not None:
            return [r for sublist in parse_relations(row[0])
                    for r in sublist[1]]
        return None


def _package_essential(package: str, release: str) -> Optional[bool]:
    from .udd import connect_udd_mirror

    conn = connect_udd_mirror()
    with conn.cursor() as cursor:
        cursor.execute(
            "SELECT (essential = 'yes') FROM packages "
            "WHERE package = %s AND release = %s",
            (package, release))

        row = cursor.fetchone()
        if row is not None:
            return row[0]
        return None


def _package_build_essential(package: str, release: str) -> bool:
    from .udd import connect_udd_mirror

    conn = connect_udd_mirror()
    with conn.cursor() as cursor:
        cursor.execute(
            "select depends from packages where package = %s and release = %s",
            ('build-essential', release))
        row = cursor.fetchone()

        build_essential = set()
        for _ws1, rel, _ws2 in parse_relations(row[0] if row else ''):
            build_essential.update([r.name for r in rel])
        return package in build_essential


def _fetch_transitions(release: str) -> Dict[str, str]:
    from .dummy_transitional import find_dummy_transitional_packages
    from .udd import connect_udd_mirror

    ret = {}
    conn = connect_udd_mirror()
    for transition in (find_dummy_transitional_packages(
            conn, release)).values():
        ret[transition.from_name] = transition.to_expr
    return ret


class PackageChecker:

    def __init__(self, release: str, build: bool):
        self.release = release
        self.build = build
        self._transitions: Optional[Dict[str, str]] = None

    def package_version(self, package: str) -> Optional[Version]:
        return _package_version(package, self.release)

    def replacement(self, package: str) -> Optional[str]:
        if self._transitions is None:
            self._transitions = _fetch_transitions(self.release)
        return self._transitions.get(package)

    def package_provides(self, package):
        return _package_provides(package, self.release)

    def is_essential(self, package: str) -> Optional[bool]:
        if self.build and _package_build_essential(package, self.release):
            return True
        return _package_essential(package, self.release)


def drop_obsolete_depends(
        entry: List[PkgRelation], checker: PackageChecker,
        keep_minimum_versions: bool = False):
    ors = []
    actions: List[Action] = []
    for pkgrel in entry:
        newrel = pkgrel
        replacement = checker.replacement(pkgrel.name)
        if replacement:
            parsed_replacement = PkgRelation.parse(replacement)
            if len(parsed_replacement) != 1:
                logging.warning(
                    'Unable to replace multi-package %r', replacement)
            else:
                newrel = parsed_replacement[0]
                if newrel in entry:
                    actions.append(DropTransition(pkgrel))
                    continue
                actions.append(ReplaceTransition(pkgrel, parsed_replacement))
        elif pkgrel.version is not None and pkgrel.name != 'debhelper':
            compat_version = checker.package_version(pkgrel.name)
            logging.debug(
                "Relation: %s. Upgrade release %s has %r ",
                pkgrel, checker.release, compat_version)
            if compat_version is not None and depends_obsolete(
                compat_version, *pkgrel.version
            ):
                # If the package is essential, we don't need to maintain a
                # dependency on it.
                if checker.is_essential(pkgrel.name):
                    actions.append(DropEssential(pkgrel))
                    return [], actions
                if not keep_minimum_versions:
                    newrel = PkgRelation.parse(pkgrel.str())[0]
                    newrel.version = None
                    actions.append(DropMinimumVersion(pkgrel))
        ors.append(newrel)
    if not actions:
        return ors, []

    deduped = []
    for rel in ors:
        if rel not in deduped:
            deduped.append(rel)

    # TODO: if dropped: Check if any ors are implied by existing other
    # dependencies
    return deduped, actions


def drop_obsolete_conflicts(checker: PackageChecker, entry: List[PkgRelation]):
    ors = []
    actions: List[Action] = []
    for pkgrel in entry:
        if pkgrel.version is not None:
            compat_version = checker.package_version(pkgrel.name)
            if compat_version is not None and conflict_obsolete(
                compat_version, *pkgrel.version
            ):
                actions.append(DropObsoleteConflict(pkgrel))
                continue
        ors.append(pkgrel)
    return ors, actions


def update_depends(
        base: Deb822Dict, field: str, checker: PackageChecker,
        keep_minimum_versions: bool = False) -> List[Action]:
    return filter_relations(
        base, field,
        lambda oldrelation: drop_obsolete_depends(
            oldrelation, checker, keep_minimum_versions=keep_minimum_versions))


def _relations_empty(rels):
    return all(not rel for _ws1, rel, _ws2 in rels)


RelationsCallback = Callable[
    [List[PkgRelation]], Tuple[List[PkgRelation], List[Action]]]


def filter_relations(
        base: Union[Deb822Dict, Dict[str, str]],
        field: str, cb: RelationsCallback) -> List[Action]:
    """Update a relations field."""
    try:
        old_contents = base[field]
    except KeyError:
        return []

    oldrelations = parse_relations(old_contents)
    newrelations = []

    all_actions: List[Action] = []
    for i, (ws1, oldrelation, ws2) in enumerate(oldrelations):
        relation, actions = cb(oldrelation)
        all_actions.extend(actions)
        if relation == oldrelation or relation:
            newrelations.append((ws1, relation, ws2))
        elif i == 0 and len(oldrelations) > 1:
            # If the first item is removed, then copy the spacing to the next
            # item
            oldrelations[1] = (ws1, oldrelations[1][1], ws2)

    if all_actions:
        if _relations_empty(newrelations):
            del base[field]
        else:
            base[field] = format_relations(newrelations)
        return all_actions
    return []


def update_conflicts(
        base: Deb822Dict, field: str, checker: PackageChecker) -> List[Action]:
    return filter_relations(
        base, field,
        lambda oldrelation: drop_obsolete_conflicts(checker, oldrelation))


def drop_old_source_relations(
        source, compat_release,
        *, keep_minimum_depends_versions: bool = False
        ) -> List[Tuple[str, List[Action], str]]:
    checker = PackageChecker(compat_release, build=True)
    ret = []
    for field in [
        "Build-Depends",
        "Build-Depends-Indep",
        "Build-Depends-Arch",
    ]:
        actions = update_depends(
            source, field, checker,
            keep_minimum_versions=keep_minimum_depends_versions)
        if actions:
            ret.append((field, actions, compat_release))
    for field in ["Build-Conflicts", "Build-Conflicts-Indep",
                  "Build-Conflicts-Arch"]:
        actions = update_conflicts(source, field, checker)
        if actions:
            ret.append((field, actions, compat_release))
    return ret


def drop_old_binary_relations(
        package_checker, binary, upgrade_release: str, *,
        keep_minimum_depends_versions: bool = False
        ) -> List[Tuple[str, List[Action], str]]:
    ret = []
    for field in ["Depends", "Suggests", "Recommends", "Pre-Depends"]:
        actions = update_depends(
            binary, field, package_checker,
            keep_minimum_versions=keep_minimum_depends_versions)
        if actions:
            ret.append((field, actions, upgrade_release))

    for field in ["Conflicts", "Replaces", "Breaks"]:
        actions = update_conflicts(binary, field, package_checker)
        if actions:
            ret.append((field, actions, upgrade_release))

    return ret


def drop_old_relations(
        editor, package_checker, compat_release: str, upgrade_release: str,
        *, keep_minimum_depends_versions: bool = False
        ) -> List[Tuple[Optional[str], List[Tuple[str, List[Action], str]]]]:
    actions: List[
        Tuple[Optional[str], List[Tuple[str, List[Action], str]]]] = []
    source_actions = []
    try:
        check_generated_file(editor.path)
    except GeneratedFile as e:
        uses_cdbs = (
            e.template_path is not None
            and guess_template_type(e.template_path) == "cdbs"
        )
    else:
        uses_cdbs = False
    if not uses_cdbs:
        source_actions.extend(
            drop_old_source_relations(
                editor.source, compat_release,
                keep_minimum_depends_versions=keep_minimum_depends_versions))
    if source_actions:
        actions.append((None, source_actions))

    for binary in editor.binaries:
        binary_actions = drop_old_binary_relations(
            package_checker, binary, upgrade_release,
            keep_minimum_depends_versions=keep_minimum_depends_versions)
        if binary_actions:
            actions.append((binary["Package"], binary_actions))

    return actions


def update_maintscripts(
        wt: WorkingTree, subpath: str, checker: PackageChecker, package: str,
        allow_reformatting: bool = False
        ) -> List[
            Tuple[str, List[Tuple[int, str, Version]]]]:
    ret = []
    for entry in os.scandir(wt.abspath(os.path.join(subpath))):
        if not (entry.name == "maintscript"
                or entry.name.endswith(".maintscript")):
            continue
        with MaintscriptEditor(
                entry.path, allow_reformatting=allow_reformatting) as editor:
            def can_drop(p, v):
                compat_version = checker.package_version(p or package)
                return compat_version is not None and compat_version > v
            removed = drop_obsolete_maintscript_entries(editor, can_drop)
            if removed:
                ret.append((os.path.join(subpath, entry.name), removed))
    return ret


def name_list(packages: List[str]) -> str:
    """Format a list of package names for use in prose.

    Args:
      packages: non-empty list of packages to format
    Returns:
      human-readable string
    """
    if not packages:
        raise ValueError(packages)
    std = list(sorted(set(packages)))
    if len(std) == 1:
        return std[0]
    return ", ".join(std[:-1]) + " and " + std[-1]


class ScrubObsoleteResult:

    control_actions: List[
        Tuple[Optional[str], List[Tuple[str, List[Action], str]]]]

    def __init__(self, specific_files, maintscript_removed, control_actions):
        self.specific_files = specific_files
        self.maintscript_removed = maintscript_removed
        self.control_actions = control_actions
        self.changelog_behaviour = None

    def __bool__(self):
        return bool(self.control_actions) or bool(self.maintscript_removed)

    def value(self) -> int:
        value = DEFAULT_VALUE_MULTIARCH_HINT
        for _para, changes in self.control_actions:
            for _field, actions, _release in changes:
                value += len(actions) * 2
        for _path, removed, _release in self.maintscript_removed:
            value += len(removed)
        return value

    def itemized(self) -> Dict[str, List[str]]:
        summary: Dict[str, List[str]] = {}
        for para, changes in self.control_actions:
            for field, actions, release in changes:
                for action in actions:
                    if para:
                        summary.setdefault(release, []).append(
                            f"{para}: {action} in {field}.")
                    else:
                        summary.setdefault(release, []).append(
                            f"{field}: {action}.")
        if self.maintscript_removed:
            total_entries = sum(
                [len(entries)
                    for name, entries, release in self.maintscript_removed])
            summary.setdefault(self.maintscript_removed[0][2], []).append(
                "Remove %d maintscript entries from %d files." % (
                    total_entries, len(self.maintscript_removed))
            )
        return summary


def _scrub_obsolete(
        wt: WorkingTree,
        debian_path: str,
        *,
        compat_release: str,
        upgrade_release: str,
        allow_reformatting: bool = True,
        keep_minimum_depends_versions: bool = False) -> ScrubObsoleteResult:
    specific_files = []
    binary_package_checker = PackageChecker(upgrade_release, build=False)
    control_path = os.path.join(debian_path, "control")
    try:
        with ControlEditor(
                wt.abspath(control_path),
                allow_reformatting=allow_reformatting) as editor:
            package = editor.source["Source"]
            control_actions = drop_old_relations(
                editor, binary_package_checker, compat_release,
                upgrade_release,
                keep_minimum_depends_versions=keep_minimum_depends_versions)
        specific_files.extend(wt.safe_relpath_files(editor.changed_files))
    except FileNotFoundError as exc:
        if wt.has_filename(os.path.join(debian_path, "debcargo.toml")):
            control_actions = []
        else:
            raise NotDebianPackage(wt.abspath(debian_path)) from exc

    maintscript_removed = []
    for path, removed in update_maintscripts(
            wt, debian_path, binary_package_checker,
            package, allow_reformatting):
        if removed:
            maintscript_removed.append((path, removed, upgrade_release))
            specific_files.append(path)

    return ScrubObsoleteResult(
        specific_files=specific_files,
        control_actions=control_actions,
        maintscript_removed=maintscript_removed,
    )


def release_aliases(name):
    from distro_info import DebianDistroInfo, UbuntuDistroInfo
    ret = []
    debian_distro_info = DebianDistroInfo()
    ubuntu_distro_info = UbuntuDistroInfo()
    FN_ALIAS_MAP = {
        debian_distro_info.stable: 'stable',
        debian_distro_info.old: 'oldstable',
        debian_distro_info.devel: 'unstable',
        ubuntu_distro_info.lts: 'lts',
        ubuntu_distro_info.stable: 'stable'
    }
    for fn, alias in FN_ALIAS_MAP.items():
        if fn() == name:
            ret.append(alias)
    if ret:
        return '(%s)' % ', '.join(ret)
    return ''


def scrub_obsolete(
        wt: WorkingTree, subpath: str,
        *,
        compat_release: str,
        upgrade_release: str,
        update_changelog=None,
        allow_reformatting: bool = False,
        keep_minimum_depends_versions: bool = False,
        transitions: Optional[Dict[str, str]] = None) -> ScrubObsoleteResult:
    """Scrub obsolete entries."""
    if control_files_in_root(wt, subpath):
        debian_path = subpath
    else:
        debian_path = os.path.join(subpath, 'debian')

    result = _scrub_obsolete(
        wt, debian_path, compat_release=compat_release,
        upgrade_release=upgrade_release, allow_reformatting=allow_reformatting,
        keep_minimum_depends_versions=keep_minimum_depends_versions)

    if not result:
        return result

    specific_files = list(result.specific_files)
    summary = result.itemized()

    changelog_path = os.path.join(debian_path, "changelog")

    if update_changelog is None:
        from .detect_gbp_dch import guess_update_changelog

        dch_guess = guess_update_changelog(wt, debian_path)
        if dch_guess:
            update_changelog = dch_guess.update_changelog
            _note_changelog_policy(update_changelog, dch_guess.explanation)
            result.changelog_behaviour = dch_guess
        else:
            # Assume we should update changelog
            update_changelog = True
            result.changelog_behaviour = None

    if update_changelog:
        lines = []
        for release, entries in summary.items():
            rev_aliases = release_aliases(release)
            lines.append(
                "Remove constraints unnecessary since %s" % release
                + ((' ' + rev_aliases) if rev_aliases else '') + ':')
            lines.extend(["+ " + line for line in entries])
        add_changelog_entry(wt, changelog_path, lines)
        specific_files.append(changelog_path)

    lines = []
    for release, entries in summary.items():
        rev_aliases = release_aliases(release)
        lines.extend(
            ["Remove constraints unnecessary since %s" % release
             + ((' ' + rev_aliases) if rev_aliases else ''), ""])
        lines.extend(["* " + line for line in entries])
    lines.extend(["", "Changes-By: deb-scrub-obsolete"])

    committer = get_committer(wt)

    with suppress(PointlessCommit):
        wt.commit(
            specific_files=specific_files,
            message="\n".join(lines),
            allow_pointless=False,
            reporter=NullCommitReporter(),
            committer=committer,
        )

    return result


def versions_dict() -> dict[str, str]:
    import debmutate

    import debian
    import lintian_brush
    return {
        "lintian-brush": lintian_brush.version_string,
        "debmutate": debmutate.version_string,
        "debian": debian.__version__,
    }


def report_okay(code: str, description: str) -> None:
    if os.environ.get('SVP_API') == '1':
        with open(os.environ['SVP_RESULT'], 'w') as f:
            json.dump({
                'result_code': code,
                'versions': versions_dict(),
                'description': description}, f)
    logging.info('%s', description)


def main():  # noqa: C901
    import argparse

    import breezy  # noqa: E402
    from breezy.errors import NotBranchError

    breezy.initialize()  # type: ignore
    import breezy.bzr  # noqa: E402
    import breezy.git  # noqa: E402
    from breezy.workspace import (
        WorkspaceDirty,
        check_clean_tree,
    )

    from . import (
        version_string,
    )
    from .config import Config

    parser = argparse.ArgumentParser(prog="deb-scrub-obsolete")
    parser.add_argument(
        "--directory",
        metavar="DIRECTORY",
        help="directory to run in",
        type=str,
        default=".",
    )
    parser.add_argument(
        "--upgrade-release",
        metavar="UPGRADE-RELEASE",
        help="Release to allow upgrading from.",
        default="oldstable",
    )
    parser.add_argument(
        '--compat-release',
        metavar='COMPAT-RELEASE',
        help="Release to allow building on.",
        default=os.environ.get('COMPAT_RELEASE'))
    parser.add_argument(
        "--no-update-changelog",
        action="store_false",
        default=None,
        dest="update_changelog",
        help="do not update the changelog",
    )
    parser.add_argument(
        "--update-changelog",
        action="store_true",
        dest="update_changelog",
        help="force updating of the changelog",
        default=None,
    )
    parser.add_argument(
        "--allow-reformatting",
        default=None,
        action="store_true",
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--keep-minimum-depends-versions",
        action="store_true",
        help="Keep minimum version dependencies, even when unnecessary")
    parser.add_argument(
        "--version", action="version", version="%(prog)s " + version_string
    )
    parser.add_argument(
        "--identity",
        help="Print user identity that would be used when committing",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--debug", help="Describe all considered changes.", action="store_true"
    )

    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO, format='%(message)s')

    try:
        wt, subpath = WorkingTree.open_containing(args.directory)
    except NotBranchError:
        logging.error(
            "No version control directory found (e.g. a .git directory).")
        return 1

    if args.identity:
        logging.info('%s', get_committer(wt))
        return 0

    with wt.lock_write():
        try:
            check_clean_tree(wt, wt.basis_tree(), subpath)
        except WorkspaceDirty:
            logging.info(
                "%s: Please commit pending changes first.", wt.basedir)
            return 1

        import distro_info
        debian_info = distro_info.DebianDistroInfo()
        upgrade_release = debian_info.codename(args.upgrade_release)

        update_changelog = args.update_changelog
        allow_reformatting = args.allow_reformatting
        if args.compat_release:
            compat_release = debian_info.codename(args.compat_release)
        else:
            compat_release = None
        try:
            cfg = Config.from_workingtree(wt, subpath)
        except FileNotFoundError:
            pass
        else:
            if update_changelog is None:
                update_changelog = cfg.update_changelog()
            if allow_reformatting is None:
                allow_reformatting = cfg.allow_reformatting()
            if compat_release is None:
                compat_release = cfg.compat_release()

        if compat_release is None:
            compat_release = debian_info.codename('oldstable')

        if upgrade_release != compat_release:
            logging.info(
                "Removing run time constraints unnecessary since %s"
                " and build time constraints unnecessary since %s",
                upgrade_release, compat_release)
        else:
            logging.info(
                "Removing run time and build time constraints unnecessary "
                "since %s", compat_release)

        if allow_reformatting is None:
            allow_reformatting = False

        if is_debcargo_package(wt, subpath):
            report_fatal(
                versions_dict(), "nothing-to-do", "Package uses debcargo")
            return 1
        elif not control_file_present(wt, subpath):
            report_fatal(
                versions_dict(), "missing-control-file",
                "Unable to find debian/control")
            return 1

        try:
            result = scrub_obsolete(
                wt, subpath, compat_release=compat_release,
                upgrade_release=upgrade_release,
                update_changelog=args.update_changelog,
                allow_reformatting=allow_reformatting,
                keep_minimum_depends_versions=(
                    args.keep_minimum_depends_versions)
            )
        except FormattingUnpreservable as e:
            report_fatal(
                versions_dict(), "formatting-unpreservable",
                "unable to preserve formatting while editing %s" % e.path,
            )
            if hasattr(e, 'diff'):  # debmutate >= 0.64
                sys.stderr.writelines(e.diff())
            return 1
        except GeneratedFile as e:
            report_fatal(
                versions_dict(), "generated-file",
                "unable to edit generated file: %r" % e
            )
            return 1
        except NotDebianPackage:
            report_fatal(
                versions_dict(), 'not-debian-package', 'Not a Debian package.')
            return 1
        except ChangeConflict as e:
            report_fatal(
                versions_dict(), 'change-conflict',
                'Generated file changes conflict: %s' % e)
            return 1
        except UddError as e:
            report_fatal(
                versions_dict(), 'udd-error',
                f'Error communicating with UDD: {e}')
            return 1

    if not result:
        report_okay("nothing-to-do", "no obsolete constraints")
        return 0

    debian_context = {}
    if result.changelog_behaviour:
        debian_context['changelog'] = result.changelog_behaviour.json()

    if svp_enabled():
        with open(os.environ["SVP_RESULT"], "w") as f:
            json.dump({
                "description": "Remove constraints unnecessary since %s."
                % upgrade_release,
                "value": result.value(),
                "debian": debian_context,
                "versions": versions_dict(),
                "context": {
                    "specific_files": result.specific_files,
                    "maintscript_removed": [
                        (name, [
                            (lineno, pkg, str(version) if version else None)
                            for (lineno, pkg, version) in entries], release)
                        for (name, entries, release)
                        in result.maintscript_removed],
                    "control_actions": [
                        (pkg, [(field, [action.json() for action in actions],
                                release)
                               for (field, actions, release) in changes])
                        for (pkg, changes) in result.control_actions],
                }
            }, f)

    logging.info("Scrub obsolete settings.")
    for lines in result.itemized().values():
        for line in lines:
            logging.info("* %s", line)

    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/python3

import sys

try:
    import psycopg2  # noqa: F401
except ModuleNotFoundError:
    sys.exit(2)

from debmutate.vcs import (
    split_vcs_url,
    unsplit_vcs_url,
)

from lintian_brush.fixer import (
    control,
    diligence,
    opinionated,
    report_result,
    warn,
)
from lintian_brush.vcs import determine_browser_url

if diligence() < 1:
    # Removing unused overrides requires pro-actively contacting UDD.
    sys.exit(0)

from lintian_brush.vcswatch import VcsWatch, VcsWatchError


def get_default_branch(url, branch=None):
    from dulwich.client import get_transport_and_path
    c, p = get_transport_and_path(url)
    result = c.fetch_pack(p, lambda rs: [], None, None)
    if branch is not None:
        ref = b'refs/headss' + branch.encode('utf-8')
    else:
        ref = b'HEAD'
    try:
        head = result.symrefs[ref]
    except KeyError:
        return None
    if not head.startswith(b'refs/heads/'):
        return head.decode('utf-8')
    return head[len(b'refs/heads/'):].decode('utf-8')


def find_branch(vcs_type, url):
    with VcsWatch() as vcswatch:
        return vcswatch.get_branch_from_url(vcs_type, vcs_git)


with control as updater:
    try:
        vcs_git = updater.source['Vcs-Git']
    except KeyError:
        pass
    else:
        repo_url, branch, subpath = split_vcs_url(vcs_git)
        try:
            new_branch = find_branch('Git', vcs_git)
        except KeyError:
            pass
        except VcsWatchError as e:
            warn('vcswatch URL unusable: %s' % e.args[0])
        else:
            if branch != new_branch:
                default_branch = get_default_branch(repo_url)
                if opinionated() or new_branch not in (default_branch, branch):
                    updater.source['Vcs-Git'] = unsplit_vcs_url(
                        repo_url, new_branch, subpath)
                    vcs_browser = determine_browser_url(
                        'Git', updater.source['Vcs-Git'])
                    if vcs_browser is not None:
                        updater.source['Vcs-Browser'] = vcs_browser
                    report_result("Set branch from vcswatch in Vcs-Git URL.")



#!/usr/bin/python3

from debmutate._rules import update_rules

from lintian_brush.fixer import LintianIssue, report_result


def update_line(line, target):
    if line.strip() == b'dh_clean -k':
        issue = LintianIssue('source', "dh-clean-k-is-deprecated", '')
        if issue.should_fix():
            issue.report_fixed()
            return b'dh_prep'
    return line


update_rules(update_line)
report_result("debian/rules: Use dh_prep rather than \"dh_clean -k\".")
